import json


class Clustering(object):
    @staticmethod
    def __convert_back_to_original_nodes_names_if_needed(communities: list) -> list:
        """
        If original nodes are int and we converted the graph to igraph, they were transformed to int. So we need to
        transform them back to int
        :return:
        """

        if (
            len(communities) > 0
            and not isinstance(communities[0], dict)
            and isinstance(list(communities[0])[0], str)
        ):
            if communities[0][0][:1] == "\\":
                to_return = []
                for com in communities:
                    to_return.append([int(x[1:]) for x in com])
                return to_return
        return communities

    def __init__(
        self,
        communities: list,
        graph: object,
        method_name: str = "",
        method_parameters: dict = None,
        overlap: bool = False,
    ):
        """
        Communities representation.

        :param communities: list of communities (community: list of nodes)
        :param method_name: algorithms discovery algorithm name
        :param overlap: boolean, whether the partition is overlapping or not
        """
        if isinstance(communities, set):
            communities = list(communities)
        communities = self.__convert_back_to_original_nodes_names_if_needed(communities)
        self.communities = sorted(communities, key=len, reverse=True)
        self.graph = graph
        self.method_name = method_name
        self.overlap = overlap
        self.method_parameters = method_parameters
        self.node_coverage = 1

    def to_json(self) -> str:
        """
        Generate a JSON representation of the algorithms object

        :return: a JSON formatted string representing the object
        """
        partition = {
            "communities": self.communities,
            "algorithm": self.method_name,
            "params": self.method_parameters,
            "overlap": self.overlap,
            "coverage": self.node_coverage,
        }

        return json.dumps(partition)

    def get_description(
        self, parameters_to_display: list = None, precision: int = 3
    ) -> str:
        """
        Return a description of the clustering, with the name of the method and its numeric parameters.

        :param parameters_to_display: parameters to display. By default, all float parameters.
        :param precision: precision used to plot parameters. default: 3
        :return: a string description of the method.
        """
        description = self.method_name

        # if no parameter name provided, display all parameters (or return directly description if none)

        if parameters_to_display is None:
            if self.method_parameters is not None:
                for p in self.method_parameters:
                    if self.method_parameters[p] is not None:
                        parameters_to_display = self.method_parameters.keys()
                if parameters_to_display is None:
                    return description
            else:
                return description

        description += "("
        # for each parameter, if it is a float, add it with the required precision.
        for p in parameters_to_display:
            if isinstance(self.method_parameters[p], float):
                description += p + ":"
                description += "{1:.{0}f}".format(precision, self.method_parameters[p])
                description += ", "
            elif isinstance(self.method_parameters[p], int):
                description += p + ":"
                description += "%s" % (self.method_parameters[p])
                description += ", "

        # handle formatting
        if description[-2:] == ", ":
            description = description[:-2]
        description += ")"
        return description



import networkx as nx
import numpy as np
import random

__author__ = "Giulio Rossetti"
__license__ = "GPL"
__email__ = "giulio.rossetti@gmail.com"
__version__ = "0.2.0"


class RDynV2(object):
    def __init__(
        self,
        size: int = 300,
        iterations: int = 5,
        avg_deg: int = 15,
        sigma: float = 0.6,
        lambdad: int = 1,
        alpha: float = 2.5,
        paction: int = 1,
        prenewal: float = 0.8,
        quality_threshold: float = 0.2,
        new_node: float = 0.0,
        del_node: float = 0.0,
        max_evts: int = 1,
    ):

        # set the network generator parameters
        self.size = size
        self.iterations = iterations
        self.avg_deg = avg_deg
        self.sigma = sigma
        self.lambdad = lambdad
        self.exponent = alpha
        self.paction = paction
        self.renewal = prenewal
        self.new_node = new_node
        self.del_node = del_node
        self.max_evts = max_evts

        # event targets
        self.communities_involved = []

        # initialize communities data structures
        self.communities = {}
        self.node_to_com = [i for i in range(0, size)]
        self.total_coms = 0
        self.performed_community_action = "START\n"
        self.quality_threshold = quality_threshold
        self.exp_node_degs = []

        # initialize the graph
        self.graph = nx.empty_graph(self.size)
        self.stable = 0

        self.it = 0
        self.count = 0

    def execute(self, simplified: bool = True) -> [object, object]:
        """
        :return:
        """
        # generate degree sequence
        self.__compute_degree_sequence()

        # generate community size dist
        exp_com_s = self.__compute_community_size_distribution()

        # assign node to community
        self.__node_to_community_initial_assignement(exp_com_s)

        # main loop (iteration)
        while self.it < self.iterations:

            # community check and event generation
            comp = nx.number_connected_components(self.graph)
            if comp <= len(self.communities):
                if self.__test_communities():
                    self.it += 1
                    yield self.graph, self.communities
                    self.__generate_event(simplified)

            # node removal
            ar = random.random()
            if ar < self.del_node:
                self.__remove_node()

            # node addition
            ar = random.random()
            if ar < self.new_node:
                self.__add_node()

            # get nodes within communities that needs to adjust
            nodes = self.__get_nodes()

            # inner loop (nodes)
            for n in nodes:

                # discard deleted nodes
                if self.node_to_com[n] == -1:
                    continue

                # check for decayed edges
                removal = self.__get_vanished_edges(n)

                # removal phase
                for n1 in removal:
                    r = random.random()

                    # edge renewal phase
                    # check for intra/inter renewal thresholds
                    if (
                        r <= self.renewal
                        and self.node_to_com[n1] == self.node_to_com[n]
                        or r > self.renewal
                        and self.node_to_com[n1] != self.node_to_com[n]
                    ):

                        # Exponential decay
                        timeout = (self.it + 1) + int(random.expovariate(self.lambdad))
                        self.graph.adj[n][n1]["d"] = timeout

                    else:
                        # edge to be removed
                        self.graph.remove_edge(n, n1)

                # expected degree reached
                if self.graph.degree(n) >= self.exp_node_degs[n]:
                    continue

                # decide if the node is active during this iteration
                action = random.random()

                # the node has not yet reached it expected degree and it acts in this round
                if self.graph.degree([n])[n] < self.exp_node_degs[n] and (
                    action <= self.paction or self.it == 0
                ):

                    com_nodes = list(self.communities[self.node_to_com[n]])

                    # probability for intra/inter community edges
                    r = random.random()

                    # check if at least sigma% of the node link are within the community
                    s = self.graph.subgraph(com_nodes)
                    d = s.degree([n])[n]  # Intra community edges

                    if r <= self.sigma and d < len(com_nodes) - 1:
                        self.__new_intra_community_edge(s, n)

                    # inter-community edges
                    elif r > self.sigma:
                        self.__new_inter_community_edge(n)

    def __new_intra_community_edge(self, s, n):

        n_neigh = set(s.neighbors(n))

        random.shuffle(list(n_neigh))
        target = None

        # selecting target node
        candidates = {
            j: (self.exp_node_degs[j] - self.graph.degree(j))
            for j in s.nodes()
            if (self.exp_node_degs[j] - self.graph.degree(j)) > 0 and j != n
        }

        if len(candidates) > 0:
            target = random.sample(list(candidates), 1)[0]

        # Interaction Exponential decay
        timeout = (self.it + 1) + int(random.expovariate(self.lambdad))

        # Edge insertion
        if target is not None and not self.graph.has_edge(n, target) and target != n:
            self.graph.add_edge(n, target, d=timeout)
            self.count += 1

    def __new_inter_community_edge(self, n):
        # randomly identifying a target community
        try:
            cid = random.sample(
                set(self.communities.keys()) - {self.node_to_com[n]}, 1
            )[0]
        except:
            return

        s = self.graph.subgraph(self.communities[cid])

        # check for available nodes within the identified community
        candidates = {
            j: (self.exp_node_degs[j] - self.graph.degree(j))
            for j in s.nodes()
            if (self.exp_node_degs[j] - self.graph.degree(j)) > 0 and j != n
        }

        # PA selection on available community nodes
        if len(candidates) > 0:
            candidatesp = list(
                np.array(list(candidates.values()), dtype="float")
                / sum(list(candidates.values()))
            )
            target = np.random.choice(list(candidates.keys()), 1, candidatesp)[0]

            if self.graph.has_node(target) and not self.graph.has_edge(n, target):
                # Interaction exponential decay
                timeout = (self.it + 1) + int(random.expovariate(self.lambdad))
                self.graph.add_edge(n, target, d=timeout)
                self.count += 1

    def __compute_degree_sequence(self):
        minv = float(self.avg_deg) / (2 ** (1 / (self.exponent - 1)))
        s = [2]
        while not nx.is_graphical(s):
            s = list(
                map(
                    int,
                    nx.utils.powerlaw_sequence(
                        self.graph.number_of_nodes(), self.exponent
                    ),
                )
            )
            x = [int(p + minv) for p in s]
        self.exp_node_degs = sorted(x)

    def __compute_community_size_distribution(self):
        min_node_degree = min(self.exp_node_degs)
        min_size_com = int(min_node_degree * self.sigma)
        s = list(
            map(
                int,
                nx.utils.powerlaw_sequence(
                    int(self.graph.number_of_nodes() / self.avg_deg), self.exponent
                ),
            )
        )
        sizes = sorted([p + min_size_com for p in s])
        while sum(sizes) > self.graph.number_of_nodes():
            for i, cs in enumerate(sizes):
                sizes[i] = cs - 1

        for c, _ in enumerate(sizes):
            self.communities[c] = []
        return sizes

    def __node_to_community_initial_assignement(self, community_sizes):
        degs = [(i, v) for i, v in enumerate(self.exp_node_degs)]
        unassigned = []

        for n in degs:
            nid, nd = n

            assigned = False
            for c, c_size in enumerate(community_sizes):
                c_taken = len(self.communities[c])

                # check if the node can be added to the community
                if float(nd) * self.sigma <= c_size and c_taken < c_size:
                    self.communities[c].append(nid)
                    assigned = True
                    break

            if not assigned:
                unassigned.append(n)

        if len(unassigned) > 0:
            for i in unassigned:
                for cid in self.communities:
                    self.communities[cid].append(i[0])
                    community_sizes[cid] += 1
                    self.exp_node_degs[i[0]] = community_sizes[cid] - 1
                    break

        ntc = []
        for cid, nodes in self.communities.items():
            for n in nodes:
                ntc.append((n, cid))

        for node_com in ntc:
            self.node_to_com[node_com[0]] = node_com[1]

    def __test_communities(self):
        mcond = 0
        for k in self.communities_involved:
            c = self.communities[k]
            if len(c) == 0:
                return False

            s = self.graph.subgraph(c)
            comps = nx.number_connected_components(s)

            if comps > 1:
                cs = nx.connected_components(s)
                i = random.sample(next(cs), 1)[0]
                j = random.sample(next(cs), 1)[0]
                timeout = (self.it + 1) + int(random.expovariate(self.lambdad))
                self.graph.add_edge(i, j, d=timeout)
                self.count += 1
                return False

            score = self.__conductance_test(k, s)
            if score > mcond:
                mcond = score
        if mcond > self.quality_threshold:
            return False
        return True

    def __conductance_test(self, comid, community):
        s_degs = community.degree()
        g_degs = self.graph.degree(community.nodes())

        # Conductance
        edge_across = 2 * sum([g_degs[n] - s_degs[n] for n in community.nodes()])
        c_nodes_total_edges = community.number_of_edges() + (2 * edge_across)

        if edge_across > 0:
            ratio = float(edge_across) / float(c_nodes_total_edges)
            if ratio > self.quality_threshold:
                self.communities_involved.append(comid)
                self.communities_involved = list(set(self.communities_involved))

                for i in community.nodes():
                    nn = list(self.graph.neighbors(i))
                    for j in nn:
                        if j not in community.nodes():
                            self.count += 1
                            self.graph.remove_edge(i, j)
                            continue
            return ratio
        return 0

    def __generate_event(self, simplified=True):

        communities_involved = []
        self.stable += 1

        options = ["M", "S"]

        evt_number = random.sample(range(1, self.max_evts + 1), 1)[0]
        evs = np.random.choice(options, evt_number, p=[0.5, 0.5], replace=True)
        chosen = []

        if len(self.communities) == 1:
            evs = "S"

        self.total_coms = len(self.communities)

        self.performed_community_action = ""

        for p in evs:

            if p == "M":
                # Generate a single merge
                if len(self.communities) == 1:
                    continue
                candidates = list(set(self.communities.keys()) - set(chosen))

                # promote merging of small communities
                cl = [len(v) for c, v in self.communities.items() if c in candidates]
                comd = 1 - np.array(cl, dtype="float") / sum(cl)
                comd /= sum(comd)

                ids = []
                try:
                    ids = np.random.choice(candidates, 2, p=list(comd), replace=False)
                except:
                    continue

                chosen.extend(ids)
                communities_involved.extend([ids[0]])

                for node in self.communities[ids[1]]:
                    self.node_to_com[node] = ids[0]

                self.communities[ids[0]].extend(self.communities[ids[1]])
                del self.communities[ids[1]]

            else:
                # Generate a single splits
                if len(self.communities) == 1:
                    continue

                candidates = list(set(self.communities.keys()) - set(chosen))

                cl = [len(v) for c, v in self.communities.items() if c in candidates]
                comd = np.array(cl, dtype="float") / sum(cl)

                try:
                    ids = np.random.choice(candidates, 1, p=list(comd), replace=False)
                except:
                    continue

                c_nodes = len(self.communities[ids[0]])

                if c_nodes > 6:
                    try:
                        size = random.sample(range(3, c_nodes - 3), 1)[0]
                        first = random.sample(self.communities[ids[0]], size)
                    except:
                        continue

                    cid = max(list(self.communities.keys())) + 1
                    chosen.extend([ids[0], cid])
                    communities_involved.extend([ids[0], cid])

                    # adjusting max degree
                    for node in first:
                        self.node_to_com[node] = cid
                        if self.exp_node_degs[node] > (len(first) - 1) * self.sigma:
                            self.exp_node_degs[node] = int(
                                (len(first) - 1) + (len(first) - 1) * (1 - self.sigma)
                            )

                    self.communities[cid] = first
                    self.communities[ids[0]] = [
                        ci for ci in self.communities[ids[0]] if ci not in first
                    ]

                    # adjusting max degree
                    for node in self.communities[ids[0]]:
                        if (
                            self.exp_node_degs[node]
                            > (len(self.communities[ids[0]]) - 1) * self.sigma
                        ):
                            self.exp_node_degs[node] = int(
                                (len(self.communities[ids[0]]) - 1)
                                + (len(self.communities[ids[0]]) - 1) * (1 - self.sigma)
                            )

        if not simplified:
            self.communities_involved = list(self.communities.keys())
        else:
            self.communities_involved = communities_involved

    def __add_node(self):
        nid = self.size
        self.graph.add_node(nid)
        cid = random.sample(list(self.communities.keys()), 1)[0]
        self.communities[cid].append(nid)
        self.node_to_com.append(cid)
        deg = random.sample(
            range(
                2,
                int(
                    (len(self.communities[cid]) - 1)
                    + (len(self.communities[cid]) - 1) * (1 - self.sigma)
                ),
            ),
            1,
        )[0]
        if deg == 0:
            deg = 1
        self.exp_node_degs.append(deg)
        self.size += 1

    def __remove_node(self):

        com_sel = [c for c, v in self.communities.items() if len(v) > 3]
        if len(com_sel) > 0:
            cid = random.sample(com_sel, 1)[0]
            s = self.graph.subgraph(self.communities[cid])
            sd = dict(s.degree)
            min_value = min(sd.values())
            candidates = [k for k in sd if sd[k] == min_value]
            nid = random.sample(candidates, 1)[0]
            eds = list(self.graph.edges([nid]))
            for e in eds:
                self.count += 1
                self.graph.remove_edge(e[0], e[1])

            self.exp_node_degs[nid] = 0
            self.node_to_com[nid] = -1
            nodes = set(self.communities[cid])
            self.communities[cid] = list(nodes - {nid})
            self.graph.remove_node(nid)

    def __get_nodes(self):
        if len(self.communities_involved) == 0:
            return self.graph.nodes()
        else:
            nodes = {}
            for cid in self.communities_involved:
                for nid in self.communities[cid]:
                    nodes[nid] = None
            return list(nodes.keys())

    def __get_vanished_edges(self, n):
        removal = []
        node_neighbors = nx.all_neighbors(self.graph, n)
        if len(self.communities) >= nx.number_connected_components(self.graph):
            for n1 in node_neighbors:
                delay = self.graph.get_edge_data(n, n1)["d"]
                if delay == self.it:
                    removal.append(n1)
        return removal


# def main():
#     import argparse
#
#     sys.stdout.write("-------------------------------------\n")
#     sys.stdout.write("               {RDyn}                \n")
#     sys.stdout.write("           Graph Generator      \n")
#     sys.stdout.write("     Handling Community Dynamics  \n")
#     sys.stdout.write("-------------------------------------\n")
#     sys.stdout.write("Author: " + __author__ + "\n")
#     sys.stdout.write("Email:  " + __email__ + "\n")
#     sys.stdout.write("------------------------------------\n")
#
#     parser = argparse.ArgumentParser()
#     parser.add_argument('nodes', type=int, help='Number of nodes', default=1000)
#     parser.add_argument('iterations', type=int, help='Number of iterations', default=1000)
#     parser.add_argument('simplified', type=bool, help='Simplified execution', default=True)
#     parser.add_argument('-d', '--avg_degree', type=int, help='Average node degree', default=15)
#     parser.add_argument('-s', '--sigma', type=float, help='Sigma', default=0.7)
#     parser.add_argument('-l', '--lbd', type=float, help='Lambda community size distribution', default=1)
#     parser.add_argument('-a', '--alpha', type=int, help='Alpha degree distribution', default=2.5)
#     parser.add_argument('-p', '--prob_action', type=float, help='Probability of node action', default=1)
#     parser.add_argument('-r', '--prob_renewal', type=float, help='Probability of edge renewal', default=0.8)
#     parser.add_argument('-q', '--quality_threshold', type=float, help='Conductance quality threshold', default=0.3)
#     parser.add_argument('-n', '--new_nodes', type=float, help='Probability of node appearance', default=0)
#     parser.add_argument('-j', '--delete_nodes', type=float, help='Probability of node vanishing', default=0)
#     parser.add_argument('-e', '--max_events', type=int, help='Max number of community events for stable iteration', default=1)
#
#     args = parser.parse_args()
#     rdyn = RDynV2(size=args.nodes, iterations=args.iterations, avg_deg=args.avg_degree,
#                 sigma=args.sigma, lambdad=args.lbd, alpha=args.alpha, paction=args.prob_action,
#                 prenewal=args.prob_renewal, quality_threshold=args.quality_threshold,
#                 new_node=args.new_nodes, del_node=args.delete_nodes, max_evts=args.max_events)
#     rdyn.execute(simplified=args.simplified)



from .edge_clustering import *
from .crisp_partition import *
from .overlapping_partition import *
from .attribute_clustering import *
from .bipartite_clustering import *
from .temporal_partition import *



from cdlib import BiNodeClustering
import networkx as nx
from cdlib.utils import convert_graph_formats
from collections import defaultdict
from cdlib.algorithms.internal.pycondor import condor_object, initial_community, brim

missing_packages = set()

try:
    import infomap as imp
except ModuleNotFoundError:
    missing_packages.add("infomap")
    imp = None

try:
    from wurlitzer import pipes
except ModuleNotFoundError:
    missing_packages.add("wurlitzer")
    pipes = None

try:
    import igraph as ig
except ModuleNotFoundError:
    ig = None

try:
    import leidenalg
except ModuleNotFoundError:
    missing_packages.add("leidenalg")
    leidenalg = None

if len(missing_packages) > 0:
    print(
        "Note: to be able to use all bipartite methods, you need to install some additional packages: ",
        missing_packages,
    )

__all__ = ["bimlpa", "CPM_Bipartite", "infomap_bipartite", "condor"]


def bimlpa(g_original: object, theta: float = 0.3, lambd: int = 7) -> BiNodeClustering:
    """
    BiMLPA is designed to detect the many-to-many correspondence community in bipartite networks using multi-label propagation algorithm.


    **Supported Graph Types**

    ========== ======== ======== =========
    Undirected Directed Weighted Bipartite
    ========== ======== ======== =========
    Yes        No       No       Yes
    ========== ======== ======== =========

    :param g_original: a networkx/igraph object
    :param theta: Label weights threshold. Default 0.3.
    :param lambd: The max number of labels. Default 7.
    :return: BiNodeClustering object


    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.algorithms.bipartite.generators.random_graph(100, 20, 0.1)
    >>> coms = algorithms.bimlpa(G)

    :References:

    Taguchi, Hibiki, Tsuyoshi Murata, and Xin Liu. "BiMLPA: Community Detection in Bipartite Networks by Multi-Label Propagation." International Conference on Network Science. Springer, Cham, 2020.

    .. note:: Reference implementation: https://github.com/hbkt/BiMLPA
    """
    from BiMLPA import BiMLPA_SqrtDeg, relabeling, output_community

    g = convert_graph_formats(g_original, nx.Graph)

    if not nx.algorithms.bipartite.is_bipartite(g):
        raise ValueError("The graph is not bipartite")

    bimlpa = BiMLPA_SqrtDeg(g, theta, lambd)
    bimlpa.start()
    relabeling(g)
    top_coms, bottom_coms = output_community(g)

    return BiNodeClustering(
        top_coms,
        bottom_coms,
        g_original,
        "BiMLPA",
        method_parameters={"theta": theta, "lambd": lambd},
    )


def CPM_Bipartite(
    g_original: object,
    resolution_parameter_01: float,
    resolution_parameter_0: float = 0,
    resolution_parameter_1: float = 0,
    degree_as_node_size: bool = False,
    seed: int = 0,
) -> BiNodeClustering:
    """
    CPM_Bipartite is the extension of CPM to bipartite graphs


    **Supported Graph Types**

    ========== ======== ======== =========
    Undirected Directed Weighted Bipartite
    ========== ======== ======== =========
    Yes        No       No       Yes
    ========== ======== ======== =========

    :param g_original: a networkx/igraph object
    :param resolution_parameter_01: Resolution parameter for in between two classes.
    :param resolution_parameter_0: Resolution parameter for class 0.
    :param resolution_parameter_1: Resolution parameter for class 1.
    :param degree_as_node_size: If ``True`` use degree as node size instead of 1, to mimic modularity
    :param seed: the random seed to be used in CPM method to keep results/partitions replicable
    :return: BiNodeClustering object

    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.algorithms.bipartite.generators.random_graph(100, 20, 0.5)
    >>> coms = algorithms.CPM_Bipartite(G, 1)

    :References:

    Barber, M. J. (2007). Modularity and community detection in bipartite networks. Physical Review E, 76(6), 066102. 10.1103/PhysRevE.76.066102

    .. note:: Reference implementation: https://leidenalg.readthedocs.io/en/stable/multiplex.html?highlight=bipartite#bipartite
    """
    global leidenalg
    if ig is None or leidenalg is None:
        try:
            import leidenalg
        except ModuleNotFoundError:
            raise ModuleNotFoundError(
                "Optional dependency not satisfied: install igraph and leidenalg to use the "
                "selected feature."
            )

    g = convert_graph_formats(g_original, ig.Graph)

    try:
        g.vs["name"]
    except:
        g.vs["name"] = [v.index for v in g.vs]

    optimiser = leidenalg.Optimiser()
    leidenalg.Optimiser.set_rng_seed(self=optimiser, value=seed)

    p_01, p_0, p_1 = leidenalg.CPMVertexPartition.Bipartite(
        g,
        resolution_parameter_01=resolution_parameter_01,
        resolution_parameter_0=resolution_parameter_0,
        resolution_parameter_1=resolution_parameter_1,
        degree_as_node_size=degree_as_node_size,
    )
    optimiser.optimise_partition_multiplex([p_01, p_0, p_1], layer_weights=[1, -1, -1])

    coms = defaultdict(list)
    for n in g.vs:
        coms[p_01.membership[n.index]].append(n.index)

    return BiNodeClustering(
        list(coms.values()),
        [],
        g_original,
        "CPM_Bipartite",
        method_parameters={
            "resolution_parameter_01": resolution_parameter_01,
            "resolution_parameter_0": resolution_parameter_0,
            "resolution_parameter_1": resolution_parameter_1,
            "degree_as_node_size": degree_as_node_size,
            "seed": seed,
        },
    )


def infomap_bipartite(g_original: object, flags: str = "") -> BiNodeClustering:
    """
    Infomap is based on ideas of information theory.
    The algorithm uses the probability flow of random walks on a bipartite network as a proxy for information flows in the real system and it decomposes the network into modules by compressing a description of the probability flow.


    **Supported Graph Types**

    ========== ======== ======== =========
    Undirected Directed Weighted Bipartite
    ========== ======== ======== =========
    Yes        Yes      Yes      Yes
    ========== ======== ======== =========

    :param g_original: a networkx/igraph object
    :param flags: str flags for Infomap
    :return: BiNodeClustering object

    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> coms = algorithms.infomap_bipartite(G)

    :References:

    Rosvall M, Bergstrom CT (2008) `Maps of random walks on complex networks reveal community structure. <https://www.pnas.org/content/105/4/1118/>`_ Proc Natl Acad SciUSA 105(4):1118–1123

    .. note:: Reference implementation: https://pypi.org/project/infomap/

    .. note:: Infomap Python API documentation: https://mapequation.github.io/infomap/python/
    """

    global imp, pipes
    if imp is None:
        try:
            import infomap as imp
        except ModuleNotFoundError:
            raise ModuleNotFoundError(
                "Optional dependency not satisfied: install infomap to use the selected feature."
            )
    if pipes is None:
        try:
            from wurlitzer import pipes
        except ModuleNotFoundError:
            raise ModuleNotFoundError(
                "Optional dependency not satisfied: install package wurlitzer to use infomap."
            )

    g = convert_graph_formats(g_original, nx.Graph)

    g1 = nx.convert_node_labels_to_integers(g, label_attribute="name")
    name_map = nx.get_node_attributes(g1, "name")
    if not nx.algorithms.bipartite.is_bipartite(g1):
        raise ValueError("The graph is not bipartite")

    X, Y = nx.algorithms.bipartite.sets(g1)
    X = {x: n for n, x in enumerate(X)}
    Y = {y: n + max(X.values()) + 1 for n, y in enumerate(Y)}
    Z = {**X, **Y}

    g1 = nx.relabel_nodes(g1, Z)
    inv_Z = {v: k for k, v in Z.items()}

    coms_to_node = defaultdict(list)

    with pipes():
        im = imp.Infomap(flags)
        im.bipartite_start_id = min(Y.keys())

        if tuple(map(int, imp.__version__.split("."))) >= (1, 7, 1):
            n_dict = {i: str(n) for i, n in enumerate(g1.nodes)}
            im.add_nodes(n_dict)
        else:
            im.add_nodes(g1.nodes)

        for source, target, data in g1.edges(data=True):
            if "weight" in data:
                im.add_link(source, target, data["weight"])
            else:
                im.add_link(source, target)
        im.run()

        for node_id, module_id in im.modules:
            node_name = name_map[inv_Z[node_id]]
            coms_to_node[module_id].append(node_name)

    coms_infomap = [list(c) for c in coms_to_node.values()]
    return BiNodeClustering(
        coms_infomap,
        [],
        g_original,
        "Infomap Bipartite",
        method_parameters={"flags": flags},
    )


def condor(g_original: object) -> BiNodeClustering:
    """
    BRIM algorithm for bipartite community structure detection.
    Works on weighted and unweighted graphs.


    **Supported Graph Types**

    ========== ======== ======== =========
    Undirected Directed Weighted Bipartite
    ========== ======== ======== =========
    Yes        No       No       Yes
    ========== ======== ======== =========

    :param g_original: a networkx/igraph object
    :return: BiNodeClustering object

    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> coms = algorithms.condor(G)

    :References:

    Platig, J., Castaldi, P. J., DeMeo, D., & Quackenbush, J. (2016). Bipartite community structure of eQTLs. PLoS computational biology, 12(9), e1005033.

    .. note:: Reference implementation: https://github.com/genisott/pycondor
    """

    g = convert_graph_formats(g_original, nx.Graph)
    net = nx.to_pandas_edgelist(g)
    co = condor_object(net)
    co = initial_community(co)
    co = brim(co)

    left = co["tar_memb"]
    right = co["reg_memb"]

    lefts = defaultdict(list)
    for index, row in left.iterrows():
        if isinstance(row["tar"], str):
            lefts[row["com"]].append(row["tar"])
        else:
            lefts[row["com"]].append(int(row["tar"]))

    rights = defaultdict(list)
    for index, row in right.iterrows():
        if isinstance(row["reg"], str):
            rights[row["com"]].append(row["reg"])
        else:
            rights[row["com"]].append(int(row["reg"]))

    return BiNodeClustering(
        list(lefts.values()),
        list(rights.values()),
        g_original,
        "Condor",
        method_parameters={},
    )



import numpy as np
import collections as co

try:
    import igraph as ig
except ModuleNotFoundError:
    ig = None
import operator
import itertools
import argparse


#############################
# Fuzzy Modularity Measures #
#############################


def nepusz_modularity(G, cover):
    raise NotImplementedError("See the CONGA 2010 paper")


def zhang_modularity(G, cover):
    raise NotImplementedError(
        """See 'Identification of overlapping algorithms structure in
        complex networks using fuzzy C-means clustering'"""
    )


def nicosia_modularity(G, cover):
    raise NotImplementedError(
        """See 'Extending the definition of
        modularity to directed graphs with overlapping communities'"""
    )


#############################
# Crisp modularity measures #
#############################


def count_communities(G, cover):
    """
    Helper for lazar_modularity.
    Returns a dict {v:count} where v is a vertex id and
    count is the number of different communities it is
    assigned to.
    """
    counts = {i.index: 0 for i in G.vs}
    for community in cover:
        for v in community:
            counts[v] += 1
    return counts


def get_weights(G):
    """
    Given a graph G, returns a list of weights. If the graph is unweighted,
    returns a list of 1s the same length as the number of edges.
    """
    try:
        # asssumes weight as an attribute name means graph is weighted.
        weights = G.es["weight"]
    except KeyError:
        # unweighted means all weights are 1.
        weights = [1 for e in G.es]
    return weights


def get_single_lazar_modularity(G, community, weights, counts):
    """
    Returns the lazar modularity of a single algorithms.
    """
    totalInternalWeight = sum(weights[G.es[e].index] for e in community)  # m_c in paper
    numVerticesInCommunity = len(community)  # V_c in paper
    numPossibleInternalEdges = numVerticesInCommunity * (numVerticesInCommunity - 1) / 2
    if numPossibleInternalEdges == 0:
        return 0
    edgeDensity = (
        totalInternalWeight / numPossibleInternalEdges / numVerticesInCommunity
    )
    interVsIntra = 0
    comm = set(community)
    for v in community:
        interVsIntraInternal = 0
        neighbors = G.neighbors(v)
        degree = len(neighbors)  # k_i in paper
        numCommunitiesWithin = counts[v]  # s_i in paper
        for n in neighbors:
            weight = weights[G.get_eid(v, n)]
            if n in comm:
                interVsIntraInternal += weight
            else:
                interVsIntraInternal -= weight
        interVsIntraInternal /= degree * numCommunitiesWithin
        interVsIntra += interVsIntraInternal
    return edgeDensity * interVsIntra


def lazar_modularity(G, cover):
    """
    Returns the crisp modularity measure as defined by Lazar et al. 2009
    Defined as the average edge density times normalized difference
    between inter and intracommunity edges for each algorithms.
    See CONGA 2010 or Lazar's paper for a precise definition.
    """
    numCommunities = len(cover)  # |C| in the paper
    totalModularity = 0  # M_c in the paper
    weights = get_weights(G)
    counts = count_communities(G, cover)
    for c in cover:
        totalModularity += get_single_lazar_modularity(G, c, weights, counts)
    averageModularity = 1 / numCommunities * totalModularity  #  M in the paper
    return averageModularity


##################################
# Classes for overlapping covers #
##################################


class CrispOverlap(object):
    """
    TODO
    """

    def __init__(
        self,
        graph,
        covers,
        modularities=None,
        optimal_count=None,
        modularity_measure="lazar",
    ):
        """
        Initializes a CrispOverlap object with the given parameters.
            Graph: The graph to which the object refers
            covers: a dict of VertexCovers, also referring to this graph, of the form {k : v}
                where k is the number of clusters and v is the vertexCluste
            modularities (optional): a dict of modularities of the form {c:m} where c is
                the number of clusters and m is the modularity.
            optimal_count (optional): A hint for the number of clusters to use.
            modularity_measure (optional): The name of the modularity function to use.
                Right now, the only choice is "lazar."
        """
        # Possibly figure out a better data structure like a merge
        # list that contains all information needed?

        # So far only know of Lazar's measure for crisp overlapping.
        self._measureDict = {"lazar": lazar_modularity}
        self._covers = covers
        self._graph = graph
        self._optimal_count = optimal_count
        self._modularities = modularities
        if modularity_measure in self._measureDict:
            self._modularity_measure = modularity_measure
        else:
            raise KeyError("Modularity measure not found.")

    def __getitem__(self, numClusters):
        """
        Returns the cover with the given number of clusters.
        """
        if not numClusters:
            raise KeyError("Number of clusters must be a positive integer.")
        return self._covers[numClusters]

    def __iter__(self):
        """
        Iterates over the covers in the list.
        """
        return (v for v in list(self._covers.values()))

    def __len__(self):
        """
        Returns the number of covers in the list.
        """
        return len(self._covers)

    def __bool__(self):
        """
        Returns True when there is at least one cover in the list.
        """
        return bool(self._covers)

    def __str__(self):
        """
        Returns a string representation of the list of covers.
        """
        return "{0} vertices in {1} possible covers.".format(
            len(self._graph.vs), len(self._covers)
        )

    def as_cover(self):
        """
        Returns the optimal cover (by modularity) from the object.
        """
        return self._covers[self.optimal_count]

    def recalculate_modularities(self):
        """
        Recalculates the modularities and optimal count using the modularity_measure.
        """
        modDict = {}
        for cover in self._covers.values():
            modDict[len(cover)] = self._measureDict[self._modularity_measure](
                self._graph, cover
            )
        self._modularities = modDict
        self._optimal_count = max(
            iter(self._modularities.items()), key=operator.itemgetter(1)
        )[0]
        return self._modularities

    @property
    def modularities(self):
        """
        Returns the a dict {c : m} where c is the number of clusters
        in the cover and m is the modularity. If modularity has not
        been calculated, it recalculates it for all covers. Otherwise,
        it returns the stored dict.
        Note: Call recalculate_modularities to recalculate the modularity.
        """
        if self._modularities:
            return self._modularities
        self._modularities = self.recalculate_modularities()
        return self._modularities

    @property
    def optimal_count(self):
        """Returns the optimal number of clusters for this dendrogram.
        If an optimal count hint was given at construction time and
        recalculate_modularities has not been called, this property simply returns the
        hint. If such a count was not given, this method calculates the optimal cover
        by maximizing the modularity along all possible covers in the object.
        Note: Call recalculate_modularities to recalculate the optimal count.
        """
        if self._optimal_count is not None:
            return self._optimal_count
        else:
            modularities = self.modularities
            self._optimal_count = max(
                list(modularities.items()), key=operator.itemgetter(1)
            )[0]
            return self._optimal_count

    def pretty_print_cover(self, numClusters, label="CONGA_index"):
        """
        Takes a cover in vertex-id form and prints it nicely
        using label as each vertex's name.
        """
        cover = self._covers[numClusters]
        # if label == 'CONGA_index':
        pp = [self._graph.vs[num] for num in [cluster for cluster in cover]]
        # else:
        #    pp = [G.vs[num][label] for num in [cluster for cluster in cover]]
        for count, comm in enumerate(pp):
            print("Community {0}:".format(count))
            for v in comm:
                print("\t {0}".format(v.index if label == "CONGA_index" else v[label]))
            print()

    def make_fuzzy(self):
        """
        TODO. see CONGA 2010
        """
        pass


#################################################################################################################################################
# Possible optimizations and notes:
#   * Calculating the pair betweennesses is the large bottleneck.
#       * However, calculation of all-pairs-shortest-paths
#          and pair betweennesses are highly parallelizable.
#   * Keep a record of splits or merges?
#       * Right now, we store a lot of redundant information with a new
#           VertexCover item for every split.


def conga(OG, calculate_modularities=None, optimal_count=None):
    """
    Defines the CONGA algorithm outlined in the Gregory 2007 paper
    (An Algorithm to Find Overlapping Community Structure in Networks)
    Returns a CrispOverlap object of all of the covers.
    """

    G = OG.copy()

    comm = G.components()

    # Just in case the original graph is disconnected
    nClusters = len(comm)

    # Store the original ids of all vertices
    G.vs["CONGA_orig"] = [i.index for i in OG.vs]
    allCovers = {nClusters: ig.VertexCover(OG)}
    while G.es:
        split = remove_edge_or_split_vertex(G)
        if split:
            comm = G.components().membership
            cover = get_cover(G, OG, comm)
            nClusters += 1
            # short circuit stuff would go here.
            allCovers[nClusters] = cover
    if calculate_modularities is None:
        calculate_modularities = "lazar"
    return CrispOverlap(
        OG,
        allCovers,
        modularity_measure=calculate_modularities,
        optimal_count=optimal_count,
    )


def remove_edge_or_split_vertex(G):
    """
    The heart of the CONGA algorithm. Decides which edge should be
    removed or which vertex should be split. Returns True if the
    modification split the graph.
    """
    # has the graph split this iteration?
    split = False
    eb = G.edge_betweenness()

    maxIndex, maxEb = max(enumerate(eb), key=operator.itemgetter(1))
    # We might be able to calculate vertex betweenness and edge
    # betweenness at the same time. The current internal is slower
    # than the builtin, though.
    vb = G.betweenness()

    # Only consider vertices with vertex betweenness >= max
    # edge betweenness. From Gregory 2007 step 3
    vi = [i for i, b in enumerate(vb) if b > maxEb]

    edge = G.es[maxIndex].tuple

    if not vi:
        split = delete_edge(G, edge)
    else:
        pb = pair_betweenness(G, vi)
        maxSplit, vNum, splitInstructions = max_split_betweenness(G, pb)
        if maxSplit > maxEb:
            split = split_vertex(G, vNum, splitInstructions[0])
        else:
            split = delete_edge(G, edge)
    return split


def get_cover(G, OG, comm):
    """
    Given the graph, the original graph, and a algorithms
    membership list, returns a vertex cover of the communities
    referring back to the original algorithms.
    """
    coverDict = co.defaultdict(list)
    for i, community in enumerate(comm):
        coverDict[community].append(int(G.vs[i]["CONGA_orig"]))
    return ig.clustering.VertexCover(OG, clusters=list(coverDict.values()))


def delete_edge(G, edge):
    """
    Given a graph G and one of its edges in tuple form, checks if the deletion
    splits the graph.
    """
    G.delete_edges(edge)
    return check_for_split(G, edge)


def check_for_split(G, edge):
    """
    Given an edge in tuple form, check if it splits the
    graph into two disjoint clusters. If so, it returns
    True. Otherwise, False.
    """
    # Possibly keep a record of splits.
    if edge[0] == edge[1]:
        return False
    return not G.edge_disjoint_paths(source=edge[0], target=edge[1])


def split_vertex(G, v, splitInstructions):
    """
    Splits the vertex v into two new vertices, each with
    edges depending on s. Returns True if the split
    divided the graph, else False.
    """
    new_index = G.vcount()
    G.add_vertex()
    G.vs[new_index]["CONGA_orig"] = G.vs[v]["CONGA_orig"]

    # adding all relevant edges to new vertex, deleting from old one.
    for partner in splitInstructions:
        G.add_edge(new_index, partner)
        G.delete_edges((v, partner))

    # check if the two new vertices are disconnected.
    return check_for_split(G, (v, new_index))


def order_tuple(toOrder):
    """
    Given a tuple (a, b), returns (a, b) if a <= b,
    else (b, a).
    """
    if toOrder[0] <= toOrder[1]:
        return toOrder
    return (toOrder[1], toOrder[0])


def update_betweenness(G, path, pair, count, relevant):
    """
    Given a shortest path in G, along with a count of paths
    that length, to determine weight, updates the edge and
    pair betweenness dicts with the path's new information.
    """
    weight = 1.0 / count
    pos = 0
    while pos < len(path) - 2:
        if path[pos + 1] in relevant:
            pair[path[pos + 1]][order_tuple((path[pos], path[pos + 2]))] += weight
        pos += 1


def pair_betweenness(G, relevant):
    """
    Returns a dictionary of the pair betweenness of all vertices in relevant.
    The structure of the returned dictionary is dic[v][(u, w)] = c, where c
    is the number of shortest paths traverse through u, v, w.
    """
    pair_betweenness = {
        vertex: {uw: 0 for uw in itertools.combinations(G.neighbors(vertex), 2)}
        for vertex in relevant
    }

    for i in G.vs:
        pathCounts = co.Counter()
        # Only find the shortest paths that we haven't already seen
        shortest_paths_from_v = G.get_all_shortest_paths(
            i, to=G.vs[i.index + 1 :]
        )  # here too. need all shortest paths. too bad.
        for path in shortest_paths_from_v:  # reads twice. Can I get it down to once?
            pathCounts[path[-1]] += 1
        for path in shortest_paths_from_v:
            update_betweenness(
                G, path, pair_betweenness, pathCounts[path[-1]], set(relevant)
            )
    return pair_betweenness


def create_clique(G, v, pb):
    """
    Given a vertex and its pair betweennesses, returns a k-clique
    representing all of its neighbors, with edge weights determined by the pair
    betweenness scores. Algorithm discussed on page 5 of the CONGA paper.
    """
    neighbors = G.neighbors(v)

    # map each neighbor to its index in the adjacency matrix
    mapping = {neigh: i for i, neigh in enumerate(neighbors)}
    n = len(neighbors)

    # Can use ints instead: (dtype=int). Only works if we use matrix_min
    # instead of mat_min.
    clique = np.zeros((n, n))

    for uw, score in pb.items():

        clique[mapping[uw[0]], mapping[uw[1]]] = score
        clique[mapping[uw[1]], mapping[uw[0]]] = score

    # Ignore any self loops if they're there. If not, this line
    # does nothing and can be removed.
    np.fill_diagonal(clique, 0)
    return clique


def max_split_betweenness(G, dic):
    """
    Given a dictionary of vertices and their pair betweenness scores, uses the greedy
    algorithm discussed in the CONGA paper to find a (hopefully) near-optimal split.
    Returns a 3-tuple (vMax, vNum, vSpl) where vMax is the max split betweenness,
    vNum is the vertex with said split betweenness, and vSpl is a list of which
    vertices are on each side of the optimal split.
    """
    vMax = 0
    # for every vertex of interest, we want to figure out the maximum score achievable
    # by splitting the vertices in various ways, and return that optimal split
    for v in dic:
        clique = create_clique(G, v, dic[v])

        # initialize a list on how we will map the neighbors to the collapsing matrix
        vMap = [[ve] for ve in G.neighbors(v)]

        # we want to keep collapsing the matrix until we have a 2x2 matrix and its
        # score. Then we want to remove index j from our vMap list and concatenate
        # it with the vMap[i]. This begins building a way of keeping track of how
        # we are splitting the vertex and its neighbors
        while clique.size > 4:
            i, j, clique = reduce_matrix(clique)
            vMap[i] += vMap.pop(j)
        if clique[0, 1] >= vMax:
            vMax = clique[0, 1]
            vNum = v
            vSpl = vMap
    return vMax, vNum, vSpl


def matrix_min(mat):
    """
    Given a symmetric matrix, find an index of the minimum value
    in the upper triangle (not including the diagonal.)
    """
    # Currently, this function is unused, as its result is
    # the same as that of mat_min, and it is not always
    # faster. Left in for reference in case mat_min becomes
    # a bottleneck.

    # find the minimum from the upper triangular matrix
    # (not including the diagonal)
    upperTri = np.triu_indices(mat.shape[0], 1)
    minDex = mat[upperTri].argmin()

    # find the index in the big matrix. TODO: do so
    # with some algebra.
    triN = mat.shape[0] - 1
    row = 0
    while minDex >= triN:
        minDex -= triN
        triN -= 1
        row += 1
    col = mat.shape[0] - triN + minDex
    return row, col


def mat_min(M):
    """
    Given a matrix, find an index of the minimum value (not including the
    diagonal).
    """
    # take a matrix we pass in, and fill the diagonal with the matrix max. This is
    # so that we don't grab any values from the diag.
    np.fill_diagonal(M, float("inf"))

    # figure out the indices of the cell with the lowest value.
    i, j = np.unravel_index(M.argmin(), M.shape)
    np.fill_diagonal(M, 0)
    return i, j


def reduce_matrix(M):
    """
    Given a matrix M, collapses the row and column of the minimum value. This is just
    an adjacency matrix way of implementing the greedy "collapse" discussed in CONGA.
    Returns the new matrix and the collapsed indices.
    """
    i, j = mat_min(M)
    # i, j = matrix_min(M)
    # add the ith row to the jth row and overwrite the ith row with those values
    M[i, :] = M[j, :] + M[i, :]

    # delete the jth row
    M = np.delete(M, (j), axis=0)

    # similarly with the columns
    M[:, i] = M[:, j] + M[:, i]
    M = np.delete(M, (j), axis=1)
    np.fill_diagonal(M, 0)  # not sure necessary.
    return i, j, M


def pretty_print_cover(G, cover, label="CONGA_index"):
    """
    Takes a cover in vertex-id form and prints it nicely
    using label as each vertex's name.
    """
    pp = [G.vs[num] for num in [cluster for cluster in cover]]
    for count, comm in enumerate(pp):
        print("Community {0}:".format(count))
        for v in comm:
            print("\t", end=" ")
            if label == "CONGA_index":
                print(v.index)
            else:
                print(v[label])
        print()


def run_demo():
    """
    Finds the communities of the Zachary graph and gets the optimal one using
    Lazar's measure of modularity. Finally, pretty-prints the optimal cover.
    """
    G = ig.Graph().Famous("Zachary").as_undirected()
    result = conga(G, calculate_modularities="lazar")
    result.pretty_print_cover(result.optimal_count, label="CONGA_index")


def main():
    parser = argparse.ArgumentParser(
        description="""Run CONGA from the command line. Mostly meant as a demo -- only prints one cover."""
    )
    parser.add_argument(
        "-m",
        "--modularity_measure",
        choices=["lazar"],
        help="""Calculate the modularities using the specified
                            modularity measure. Currently only supports lazar.""",
    )
    parser.add_argument(
        "-n",
        "--num_clusters",
        type=int,
        help="""Specify the number of clusters to use.""",
    )
    parser.add_argument(
        "-d",
        "--demo",
        action="store_true",
        help="""Run a demo with the famous Zachary's Karate Club data set. Overrides all other options.""",
    )
    parser.add_argument(
        "-l",
        "--label",
        default="CONGA_index",
        nargs="?",
        const="label",
        help="""Choose which attribute of the graph to print.
                            When this option is present with no parameters, defaults to 'label'. When the option is not
                            present, defaults to the index.""",
    )
    parser.add_argument(
        "file", nargs="?", help="""The path to the file in igraph readable format."""
    )
    args = parser.parse_args()

    if args.demo:
        run_demo()
        return
    if not args.file:
        print("conga.py: error: no file specified.\n")
        print(parser.parse_args("-h".split()))
        return

    # only works for undirected
    G = ig.read(args.file).as_undirected()
    result = conga(
        G,
        calculate_modularities=args.modularity_measure,
        optimal_count=args.num_clusters,
    )
    result.pretty_print_cover(result.optimal_count, label=args.label)


def Conga_(graph, number_communities=0):
    """

    :param graph:
    :param number_communities:
    :return:
    """

    result = conga(graph)

    if number_communities == 0:
        cover = result._covers[result.optimal_count]
        number_communities = result.optimal_count
    else:
        cover = result._covers[number_communities]

    list_communities = []
    for i in range(0, number_communities):
        list_communities.append(cover._clusters[i])

    return list_communities



import networkx as nx
import numpy as np
import pandas as pd

__all__ = ["ga_community_detection"]


def ga_community_detection(graph, population=300, generation=30, r=1.5):
    Adj = nx.adjacency_matrix(graph)
    nodes_length = graph.number_of_nodes()

    d = {"chrom": [__generate_chrom(nodes_length, Adj) for n in range(population)]}
    dframe = pd.DataFrame(data=d)
    dframe["subsets"] = dframe["chrom"].apply(__find_subsets)
    dframe["community_score"] = dframe.apply(
        lambda x: __community_score(x["chrom"], x["subsets"], r, Adj), axis=1
    )

    gen = 0
    population_count = population
    while gen < generation:
        for i in range(int(np.floor(population / 10))):
            elites = dframe.sort_values("community_score", ascending=True)[
                int(np.floor(population / 10)) :
            ]
            p1 = __roulette_selection(elites)
            p2 = __roulette_selection(elites)
            child = __uniform_crossover(dframe["chrom"][p1], dframe["chrom"][p2], 0.8)
            if len(child) == 0:
                continue
            child = __mutation(child, Adj, 0.2)
            child_subsets = __find_subsets(child)
            child_cs = __community_score(child, child_subsets, r, Adj)
            dframe.loc[population_count] = [child, child_subsets, child_cs]
            population_count += 1
        dfsorted = dframe.sort_values("community_score", ascending=False)
        to_drop = dfsorted.index[population:]
        dframe.drop(to_drop, inplace=True)
        gen += 1
    sorted_df = dframe.sort_values("community_score", ascending=False).index[0]

    nodes_subsets = dframe["subsets"][sorted_df]
    nodes_list = list(graph.nodes())
    result = []
    for subs in nodes_subsets:
        subset = []
        for n in subs:
            subset.append(nodes_list[n])
        result.append(subset)
    return result


def __generate_chrom(nodes_length, Adj):
    chrom = np.array([], dtype=int)
    for x in range(nodes_length):
        rand = np.random.randint(0, nodes_length)
        while Adj[x, rand] != 1:
            rand = np.random.randint(0, nodes_length)
        chrom = np.append(chrom, rand)
    return chrom


def __merge_subsets(sub):
    arr = []
    to_skip = []
    for s in range(len(sub)):
        if sub[s] not in to_skip:
            new = sub[s]
            for x in sub:
                if sub[s] & x:
                    new = new | x
                    to_skip.append(x)
            arr.append(new)
    return arr


def __find_subsets(chrom):
    sub = [{x, chrom[x]} for x in range(len(chrom))]
    result = sub
    i = 0
    while i < len(result):
        candidate = __merge_subsets(result)
        if candidate != result:
            result = candidate
        else:
            break
        # result = candidate
        i += 1
    return result


def __community_score(chrom, subsets, r, Adj):
    matrix = Adj.toarray()
    CS = 0
    for s in subsets:
        submatrix = np.zeros((len(chrom), len(chrom)), dtype=int)
        for i in s:
            for j in s:
                submatrix[i][j] = matrix[i][j]
        M = 0
        v = 0
        for row in list(s):
            row_mean = np.sum(submatrix[row]) / len(s)
            v += np.sum(submatrix[row])
            M += (row_mean ** r) / len(s)
        CS += M * v
    return CS


def __roulette_selection(df_elites):
    prob = np.random.random_sample()
    sum_cs = np.sum(df_elites["community_score"])
    x = 0
    chosen = 0
    for i in df_elites.index:
        x += df_elites["community_score"][i]
        X = x / sum_cs
        if prob < X:
            chosen = i
            break
    return chosen


def __uniform_crossover(parent_1, parent_2, crossover_rate):
    if np.random.random_sample() < crossover_rate:
        length = len(parent_1)
        mask = np.random.randint(2, size=length)
        child = np.zeros(length, dtype=int)
        for i in range(len(mask)):
            if mask[i] == 1:
                child[i] = parent_1[i]
            else:
                child[i] = parent_2[i]
        return child
    else:
        return np.array([])


def __mutation(chrom, Adj, mutation_rate):
    if np.random.random_sample() < mutation_rate:
        chrom = chrom
        neighbor = []
        while len(neighbor) < 2:
            mutant = np.random.randint(1, len(chrom))
            row = Adj[mutant].toarray()[0]
            neighbor = [i for i in range(len(row)) if row[i] == 1]
            if len(neighbor) > 1:
                neighbor.remove(chrom[mutant])
                to_change = int(np.floor(np.random.random_sample() * (len(neighbor))))
                chrom[mutant] = neighbor[to_change]
                neighbor.append(chrom[mutant])
    return chrom



import networkx as nx
from cdlib.utils import convert_graph_formats
from collections import namedtuple
import numpy as np
import scipy
from cdlib.evaluation.internal.link_modularity import cal_modularity
import Eva
from typing import Callable
from collections import defaultdict

__all__ = [
    "FitnessResult",
    "link_modularity",
    "normalized_cut",
    "internal_edge_density",
    "average_internal_degree",
    "fraction_over_median_degree",
    "expansion",
    "cut_ratio",
    "edges_inside",
    "flake_odf",
    "avg_odf",
    "max_odf",
    "triangle_participation_ratio",
    "modularity_density",
    "z_modularity",
    "erdos_renyi_modularity",
    "newman_girvan_modularity",
    "significance",
    "surprise",
    "conductance",
    "size",
    "avg_embeddedness",
    "scaled_density",
    "avg_distance",
    "hub_dominance",
    "avg_transitivity",
    "purity",
    "modularity_overlap",
]

FitnessResult = namedtuple("FitnessResult", "min max score std")
FitnessResult.__new__.__defaults__ = (None,) * len(FitnessResult._fields)


def __median(lst: list) -> int:
    return np.median(np.array(lst))


def __out_degree_fraction(g: nx.Graph, coms: list) -> list:
    nds = []
    for n in coms:
        nds.append(g.degree(n) - coms.degree(n))
    return nds


def __quality_indexes(
    graph: nx.Graph,
    communities: object,
    scoring_function: Callable[[object, object], float],
    summary: bool = True,
) -> object:
    """

    :param graph: NetworkX/igraph graph
    :param communities: NodeClustering object
    :param summary: boolean. If **True** it is returned an aggregated score for the partition is returned, otherwise individual-communitys ones. Default **True**.
    :return: If **summary==True** a FitnessResult object, otherwise a list of floats.

    """

    graph = convert_graph_formats(graph, nx.Graph)
    values = []
    for com in communities.communities:
        community = nx.subgraph(graph, com)
        try:
            values.append(scoring_function(graph, community))
        except:
            values.append(None)

    if summary:
        values = [v for v in values if v is not None]

        if len(values) > 0:
            return FitnessResult(
                min=min(values),
                max=max(values),
                score=np.mean(values),
                std=np.std(values),
            )
        else:
            return FitnessResult(min=None, max=None, score=None, std=None)
    return values


def size(graph: nx.Graph, communities: object, **kwargs: dict) -> object:
    """Size is the number of nodes in the community

    :param graph: a networkx/igraph object
    :param communities: NodeClustering object
    :param summary: boolean. If **True** it is returned an aggregated score for the partition is returned, otherwise individual-community ones. Default **True**.
    :return: If **summary==True** a FitnessResult object, otherwise a list of floats.

    Example:

    >>> from cdlib.algorithms import louvain
    >>> from cdlib import evaluation
    >>> g = nx.karate_club_graph()
    >>> communities = louvain(g)
    >>> sz = evaluation.size(g,communities)
    """

    return __quality_indexes(graph, communities, lambda g, com: len(com), **kwargs)


def scaled_density(graph: nx.Graph, communities: object, **kwargs: dict) -> object:
    """Scaled density.

    The scaled density of a community is defined as the ratio of the community density w.r.t. the complete graph density.

    :param graph: a networkx/igraph object
    :param communities: NodeClustering object
    :param summary: boolean. If **True** it is returned an aggregated score for the partition is returned, otherwise individual-community ones. Default **True**.
    :return: If **summary==True** a FitnessResult object, otherwise a list of floats.

    Example:

    >>> from cdlib.algorithms import louvain
    >>> from cdlib import evaluation
    >>> g = nx.karate_club_graph()
    >>> communities = louvain(g)
    >>> scd = evaluation.scaled_density(g,communities)
    """

    return __quality_indexes(
        graph,
        communities,
        lambda graph, coms: nx.density(nx.subgraph(graph, coms)) / nx.density(graph),
        **kwargs
    )


def avg_distance(graph: nx.Graph, communities: object, **kwargs: dict) -> object:
    """Average distance.

    The average distance of a community is defined average path length across all possible pair of nodes composing it.

    :param graph: a networkx/igraph object
    :param communities: NodeClustering object
    :param summary: boolean. If **True** it is returned an aggregated score for the partition is returned, otherwise individual-community ones. Default **True**.
    :return: If **summary==True** a FitnessResult object, otherwise a list of floats.

    Example:

    >>> from cdlib.algorithms import louvain
    >>> from cdlib import evaluation
    >>> g = nx.karate_club_graph()
    >>> communities = louvain(g)
    >>> scd = evaluation.avg_distance(g,communities)
    """

    return __quality_indexes(
        graph,
        communities,
        lambda graph, coms: nx.average_shortest_path_length(nx.subgraph(graph, coms)),
        **kwargs
    )


def hub_dominance(graph: nx.Graph, communities: object, **kwargs: dict) -> object:
    """Hub dominance.

    The hub dominance of a community is defined as the ratio of the degree of its most connected node w.r.t. the theoretically maximal degree within the community.

    :param graph: a networkx/igraph object
    :param communities: NodeClustering object
    :param summary: boolean. If **True** it is returned an aggregated score for the partition is returned, otherwise individual-community ones. Default **True**.
    :return: If **summary==True** a FitnessResult object, otherwise a list of floats.

    Example:

    >>> from cdlib.algorithms import louvain
    >>> from cdlib import evaluation
    >>> g = nx.karate_club_graph()
    >>> communities = louvain(g)
    >>> scd = evaluation.hub_dominance(g,communities)
    """

    return __quality_indexes(
        graph,
        communities,
        lambda graph, coms: max(
            [x[1] for x in list(nx.degree(nx.subgraph(graph, coms)))]
        )
        / (len(coms) - 1),
        **kwargs
    )


def avg_transitivity(graph: nx.Graph, communities: object, **kwargs: dict) -> object:
    """Average transitivity.

    The average transitivity of a community is defined the as the average clustering coefficient of its nodes w.r.t. their connection within the community itself.

    :param graph: a networkx/igraph object
    :param communities: NodeClustering object
    :param summary: boolean. If **True** it is returned an aggregated score for the partition is returned, otherwise individual-community ones. Default **True**.
    :return: If **summary==True** a FitnessResult object, otherwise a list of floats.

    Example:

    >>> from cdlib.algorithms import louvain
    >>> from cdlib import evaluation
    >>> g = nx.karate_club_graph()
    >>> communities = louvain(g)
    >>> scd = evaluation.avg_transitivity(g,communities)
    """

    return __quality_indexes(
        graph,
        communities,
        lambda graph, coms: nx.average_clustering(nx.subgraph(graph, coms)),
        **kwargs
    )


def avg_embeddedness(graph: nx.Graph, communities: object, **kwargs: dict) -> object:
    """Average embeddedness of nodes within the community.

    The embeddedness of a node n w.r.t. a community C is the ratio of its degree within the community and its overall degree.

    .. math:: emb(n,C) = \\frac{k_n^C}{k_n}

    The average embeddedness of a community C is:

    .. math:: avg_embd(c) = \\frac{1}{|C|} \sum_{i \in C} \\frac{k_n^C}{k_n}

    :param graph: a networkx/igraph object
    :param communities: NodeClustering object
    :param summary: boolean. If **True** it is returned an aggregated score for the partition is returned, otherwise individual-community ones. Default **True**.
    :return: If **summary==True** a FitnessResult object, otherwise a list of floats.

    Example:

    >>> from cdlib.algorithms import louvain
    >>> from cdlib import evaluation
    >>> g = nx.karate_club_graph()
    >>> communities = louvain(g)
    >>> ave = evaluation.avg_embeddedness(g,communities)

    :References:


    """

    return __quality_indexes(
        graph,
        communities,
        lambda g, com: np.mean(
            [float(nx.degree(nx.subgraph(g, com))[n]) / nx.degree(g)[n] for n in com]
        ),
        **kwargs
    )


def normalized_cut(graph: nx.Graph, community: object, summary: bool = True) -> object:
    """Normalized variant of the Cut-Ratio

    .. math:: f(S) = \\frac{c_S}{2m_S+c_S} + \\frac{c_S}{2(m−m_S )+c_S}

    where :math:`m` is the number of graph edges, :math:`m_S` is the number of community internal edges and :math:`c_S` is the number of community nodes.

    :param graph: a networkx/igraph object
    :param community: NodeClustering object
    :param summary: boolean. If **True** it is returned an aggregated score for the partition is returned, otherwise individual-community ones. Default **True**.
    :return: If **summary==True** a FitnessResult object, otherwise a list of floats.


    Example:

    >>> from cdlib.algorithms import louvain
    >>> from cdlib import evaluation
    >>> g = nx.karate_club_graph()
    >>> communities = louvain(g)
    >>> mod = evaluation.normalized_cut(g,communities)

    :References:

    1.Shi, J., Malik, J.: Normalized cuts and image segmentation. Departmental Papers (CIS), 107 (2000)
    """

    graph = convert_graph_formats(graph, nx.Graph)
    values = []
    for com in community.communities:
        coms = nx.subgraph(graph, com)

        ms = len(coms.edges())
        edges_outside = 0
        for n in coms.nodes():
            neighbors = graph.neighbors(n)
            for n1 in neighbors:
                if n1 not in coms:
                    edges_outside += 1
        try:
            ratio = (float(edges_outside) / ((2 * ms) + edges_outside)) + float(
                edges_outside
            ) / (2 * (len(graph.edges()) - ms) + edges_outside)
        except:
            ratio = 0
        values.append(ratio)

    if summary:
        return FitnessResult(
            min=min(values), max=max(values), score=np.mean(values), std=np.std(values)
        )
    return values


def internal_edge_density(
    graph: nx.Graph, community: object, summary: bool = True
) -> object:
    """The internal density of the community set.

     .. math:: f(S) = \\frac{m_S}{n_S(n_S−1)/2}

     where :math:`m_S` is the number of community internal edges and :math:`n_S` is the number of community nodes.

    :param graph: a networkx/igraph object
    :param community: NodeClustering object
    :param summary: boolean. If **True** it is returned an aggregated score for the partition is returned, otherwise individual-community ones. Default **True**.
    :return: If **summary==True** a FitnessResult object, otherwise a list of floats.

    Example:

    >>> from cdlib.algorithms import louvain
    >>> from cdlib import evaluation
    >>> g = nx.karate_club_graph()
    >>> communities = louvain(g)
    >>> mod = evaluation.internal_edge_density(g,communities)


    :References:

    1. Radicchi, F., Castellano, C., Cecconi, F., Loreto, V., & Parisi, D. (2004). Defining and identifying communities in networks. Proceedings of the National Academy of Sciences, 101(9), 2658-2663.
    """
    graph = convert_graph_formats(graph, nx.Graph)
    values = []
    for com in community.communities:
        coms = nx.subgraph(graph, com)

        ms = len(coms.edges())
        ns = len(coms.nodes())
        try:
            internal_density = float(ms) / (float(ns * (ns - 1)) / 2)
        except:
            internal_density = 0
        values.append(internal_density)

    if summary:
        return FitnessResult(
            min=min(values), max=max(values), score=np.mean(values), std=np.std(values)
        )
    return values


def average_internal_degree(
    graph: nx.Graph, community: object, summary: bool = True
) -> object:
    """The average internal degree of the community set.

    .. math:: f(S) = \\frac{2m_S}{n_S}

     where :math:`m_S` is the number of community internal edges and :math:`n_S` is the number of community nodes.

    :param graph: a networkx/igraph object
    :param community: NodeClustering object
    :param summary: boolean. If **True** it is returned an aggregated score for the partition is returned, otherwise individual-community ones. Default **True**.
    :return: If **summary==True** a FitnessResult object, otherwise a list of floats.

    Example:

    >>> from cdlib.algorithms import louvain
    >>> from cdlib import evaluation
    >>> g = nx.karate_club_graph()
    >>> communities = louvain(g)
    >>> mod = evaluation.average_internal_degree(g,communities)

    :References:

    1. Radicchi, F., Castellano, C., Cecconi, F., Loreto, V., & Parisi, D. (2004). Defining and identifying communities in networks. Proceedings of the National Academy of Sciences, 101(9), 2658-2663.
    """

    graph = convert_graph_formats(graph, nx.Graph)
    values = []
    for com in community.communities:
        coms = nx.subgraph(graph, com)

        ms = len(coms.edges())
        ns = len(coms.nodes())
        try:
            avg_id = float(2 * ms) / ns
        except:
            avg_id = 0
        values.append(avg_id)

    if summary:
        return FitnessResult(
            min=min(values), max=max(values), score=np.mean(values), std=np.std(values)
        )
    return values


def fraction_over_median_degree(
    graph: nx.Graph, community: object, summary: bool = True
) -> object:
    """Fraction of community nodes of having internal degree higher than the median degree value.

    .. math:: f(S) = \\frac{|\{u: u \\in S,| \{(u,v): v \\in S\}| > d_m\}| }{n_S}


    where :math:`d_m` is the internal degree median value

    :param graph: a networkx/igraph object
    :param community: NodeClustering object
    :param summary: boolean. If **True** it is returned an aggregated score for the partition is returned, otherwise individual-community ones. Default **True**.
    :return: If **summary==True** a FitnessResult object, otherwise a list of floats.

    Example:

    >>> from cdlib.algorithms import louvain
    >>> from cdlib import evaluation
    >>> g = nx.karate_club_graph()
    >>> communities = louvain(g)
    >>> mod = evaluation.fraction_over_median_degree(g,communities)

    :References:

    1. Yang, J., Leskovec, J.: Defining and evaluating network communities based on ground-truth. Knowledge and Information Systems 42(1), 181–213 (2015)
    """

    graph = convert_graph_formats(graph, nx.Graph)
    values = []
    for com in community.communities:
        coms = nx.subgraph(graph, com)

        ns = coms.number_of_nodes()
        degs = coms.degree()

        med = __median([d[1] for d in degs])
        above_med = len([d[0] for d in degs if d[1] > med])
        try:
            ratio = float(above_med) / ns
        except:
            ratio = 0
        values.append(ratio)

    if summary:
        return FitnessResult(
            min=min(values), max=max(values), score=np.mean(values), std=np.std(values)
        )
    return values


def expansion(graph: nx.Graph, community: object, summary: bool = True) -> object:
    """Number of edges per community node that point outside the cluster.

    .. math:: f(S) = \\frac{c_S}{n_S}

    where :math:`c_S` is the number of edges on the community boundary, :math:`n_S` is the number of community nodes.

    :param graph: a networkx/igraph object
    :param community: NodeClustering object
    :param summary: boolean. If **True** it is returned an aggregated score for the partition is returned, otherwise individual-community ones. Default **True**.
    :return: If **summary==True** a FitnessResult object, otherwise a list of floats.

    Example:

    >>> from cdlib.algorithms import louvain
    >>> from cdlib import evaluation
    >>> g = nx.karate_club_graph()
    >>> communities = louvain(g)
    >>> mod = evaluation.expansion(g,communities)

    :References:

    1. Radicchi, F., Castellano, C., Cecconi, F., Loreto, V., & Parisi, D. (2004). Defining and identifying communities in networks. Proceedings of the National Academy of Sciences, 101(9), 2658-2663.
    """

    graph = convert_graph_formats(graph, nx.Graph)
    values = []
    for com in community.communities:
        coms = nx.subgraph(graph, com)

        ns = len(coms.nodes())
        edges_outside = 0
        for n in coms.nodes():
            neighbors = graph.neighbors(n)
            for n1 in neighbors:
                if n1 not in coms:
                    edges_outside += 1
        try:
            exp = float(edges_outside) / ns
        except:
            exp = 0
        values.append(exp)

    if summary:
        return FitnessResult(
            min=min(values), max=max(values), score=np.mean(values), std=np.std(values)
        )
    return values


def cut_ratio(graph: nx.Graph, community: object, summary: bool = True) -> object:
    """Fraction of existing edges (out of all possible edges) leaving the community.

    ..math:: f(S) = \\frac{c_S}{n_S (n − n_S)}

    where :math:`c_S` is the number of community nodes and, :math:`n_S` is the number of edges on the community boundary

    :param graph: a networkx/igraph object
    :param community: NodeClustering object
    :param summary: boolean. If **True** it is returned an aggregated score for the partition is returned, otherwise individual-community ones. Default **True**.
    :return: If **summary==True** a FitnessResult object, otherwise a list of floats.

    Example:

    >>> from cdlib.algorithms import louvain
    >>> from cdlib import evaluation
    >>> g = nx.karate_club_graph()
    >>> communities = louvain(g)
    >>> mod = evaluation.cut_ratio(g,communities)

    :References:

    1. Fortunato, S.: Community detection in graphs. Physics reports 486(3-5), 75–174 (2010)
    """

    graph = convert_graph_formats(graph, nx.Graph)
    values = []
    for com in community.communities:
        coms = nx.subgraph(graph, com)

        ns = len(coms.nodes())
        edges_outside = 0
        for n in coms.nodes():
            neighbors = graph.neighbors(n)
            for n1 in neighbors:
                if n1 not in coms:
                    edges_outside += 1
        try:
            ratio = float(edges_outside) / (ns * (len(graph.nodes()) - ns))
        except:
            ratio = 0
        values.append(ratio)

    if summary:
        return FitnessResult(
            min=min(values), max=max(values), score=np.mean(values), std=np.std(values)
        )
    return values


def edges_inside(graph: nx.Graph, community: object, summary: bool = True) -> object:
    """Number of edges internal to the community.

    :param graph: a networkx/igraph object
    :param community: NodeClustering object
    :param summary: boolean. If **True** it is returned an aggregated score for the partition is returned, otherwise individual-community ones. Default **True**.
    :return: If **summary==True** a FitnessResult object, otherwise a list of floats.

    Example:

    >>> from cdlib.algorithms import louvain
    >>> from cdlib import evaluation
    >>> g = nx.karate_club_graph()
    >>> communities = louvain(g)
    >>> mod = evaluation.edges_inside(g,communities)

    :References:

    1. Radicchi, F., Castellano, C., Cecconi, F., Loreto, V., & Parisi, D. (2004). Defining and identifying communities in networks. Proceedings of the National Academy of Sciences, 101(9), 2658-2663.
    """

    graph = convert_graph_formats(graph, nx.Graph)
    values = []
    for com in community.communities:
        coms = nx.subgraph(graph, com)
        values.append(coms.number_of_edges())

    if summary:
        return FitnessResult(
            min=min(values), max=max(values), score=np.mean(values), std=np.std(values)
        )
    return values


def conductance(graph: nx.Graph, community: object, summary: bool = True) -> object:
    """Fraction of total edge volume that points outside the community.

    .. math:: f(S) = \\frac{c_S}{2 m_S+c_S}

    where :math:`c_S` is the number of community nodes and, :math:`m_S` is the number of community edges

    :param graph: a networkx/igraph object
    :param community: NodeClustering object
    :param summary: boolean. If **True** it is returned an aggregated score for the partition is returned, otherwise individual-community ones. Default **True**.
    :return: If **summary==True** a FitnessResult object, otherwise a list of floats.

    Example:

    >>> from cdlib.algorithms import louvain
    >>> from cdlib import evaluation
    >>> g = nx.karate_club_graph()
    >>> communities = louvain(g)
    >>> mod = evaluation.conductance(g,communities)

    :References:

    1.Shi, J., Malik, J.: Normalized cuts and image segmentation. Departmental Papers (CIS), 107 (2000)
    """

    graph = convert_graph_formats(graph, nx.Graph)
    values = []
    for com in community.communities:
        coms = nx.subgraph(graph, com)

        ms = len(coms.edges())
        edges_outside = 0
        for n in coms.nodes():
            neighbors = graph.neighbors(n)
            for n1 in neighbors:
                if n1 not in coms:
                    edges_outside += 1
        try:
            ratio = float(edges_outside) / ((2 * ms) + edges_outside)
        except:
            ratio = 0
        values.append(ratio)

    if summary:
        return FitnessResult(
            min=min(values), max=max(values), score=np.mean(values), std=np.std(values)
        )
    return values


def max_odf(graph: nx.Graph, community: object, summary: bool = True) -> object:
    """Maximum fraction of edges of a node of a community that point outside the community itself.

    .. math:: max_{u \\in S} \\frac{|\{(u,v)\\in E: v \\not\\in S\}|}{d(u)}

    where :math:`E` is the graph edge set, :math:`v` is a node in :math:`S` and :math:`d(u)` is the degree of :math:`u`

    :param graph: a networkx/igraph object
    :param community: NodeClustering object
    :param summary: boolean. If **True** it is returned an aggregated score for the partition is returned, otherwise individual-community ones. Default **True**.
    :return: If **summary==True** a FitnessResult object, otherwise a list of floats.

    Example:

    >>> from cdlib.algorithms import louvain
    >>> from cdlib import evaluation
    >>> g = nx.karate_club_graph()
    >>> communities = louvain(g)
    >>> mod = evaluation.max_odf(g,communities)

    :References:

    1. Flake, G.W., Lawrence, S., Giles, C.L., et al.: Efficient identification of web communities. In: KDD, vol. 2000, pp. 150–160 (2000)
    """

    graph = convert_graph_formats(graph, nx.Graph)
    values = []
    for com in community.communities:
        coms = nx.subgraph(graph, com)
        values.append(max(__out_degree_fraction(graph, coms)))

    if summary:
        return FitnessResult(
            min=min(values), max=max(values), score=np.mean(values), std=np.std(values)
        )
    return values


def avg_odf(graph: nx.Graph, community: object, summary: bool = True) -> object:
    """Average fraction of edges of a node of a community that point outside the community itself.

    .. math:: \\frac{1}{n_S} \\sum_{u \\in S} \\frac{|\{(u,v)\\in E: v \\not\\in S\}|}{d(u)}

    where :math:`E` is the graph edge set, :math:`v` is a node in :math:`S`, :math:`d(u)` is the degree of :math:`u` and :math:`n_S` is the set of community nodes.

    :param graph: a networkx/igraph object
    :param community: NodeClustering object
    :param summary: boolean. If **True** it is returned an aggregated score for the partition is returned, otherwise individual-community ones. Default **True**.
    :return: If **summary==True** a FitnessResult object, otherwise a list of floats.

    Example:

    >>> from cdlib.algorithms import louvain
    >>> from cdlib import evaluation
    >>> g = nx.karate_club_graph()
    >>> communities = louvain(g)
    >>> mod = evaluation.avg_odf(g,communities)

    :References:

    1. Flake, G.W., Lawrence, S., Giles, C.L., et al.: Efficient identification of web communities. In: KDD, vol. 2000, pp. 150–160 (2000)
    """

    graph = convert_graph_formats(graph, nx.Graph)
    values = []
    for com in community.communities:
        coms = nx.subgraph(graph, com)
        values.append(float(sum(__out_degree_fraction(graph, coms))) / len(coms))

    if summary:
        return FitnessResult(
            min=min(values), max=max(values), score=np.mean(values), std=np.std(values)
        )
    return values


def flake_odf(graph: nx.Graph, community: object, summary: bool = True) -> object:
    """Fraction of nodes in S that have fewer edges pointing inside than to the outside of the community.

    .. math:: f(S) = \\frac{| \{ u:u \in S,| \{(u,v) \in E: v \in S \}| < d(u)/2 \}|}{n_S}

    where :math:`E` is the graph edge set, :math:`v` is a node in :math:`S`, :math:`d(u)` is the degree of :math:`u` and :math:`n_S` is the set of community nodes.

    :param graph: a networkx/igraph object
    :param community: NodeClustering object
    :param summary: boolean. If **True** it is returned an aggregated score for the partition is returned, otherwise individual-community ones. Default **True**.
    :return: If **summary==True** a FitnessResult object, otherwise a list of floats.

    Example:

    >>> from cdlib.algorithms import louvain
    >>> from cdlib import evaluation
    >>> g = nx.karate_club_graph()
    >>> communities = louvain(g)
    >>> mod = evaluation.flake_odf(g,communities)

    :References:

    1. Flake, G.W., Lawrence, S., Giles, C.L., et al.: Efficient identification of web communities. In: KDD, vol. 2000, pp. 150–160 (2000)
    """

    graph = convert_graph_formats(graph, nx.Graph)
    values = []
    for com in community.communities:
        coms = nx.subgraph(graph, com)
        df = 0
        for n in coms:
            fr = coms.degree(n) - (graph.degree(n) - coms.degree(n))
            if fr < 0:
                df += 1
        score = float(df) / len(coms)
        values.append(score)

    if summary:
        return FitnessResult(
            min=min(values), max=max(values), score=np.mean(values), std=np.std(values)
        )
    return values


def triangle_participation_ratio(
    graph: nx.Graph, community: object, summary: bool = True
) -> object:
    """Fraction of community nodes that belong to a triad.

    .. math:: f(S) = \\frac{ | \{ u: u \in S,\{(v,w):v, w \in S,(u,v) \in E,(u,w) \in E,(v,w) \in E \} \\not = \\emptyset \} |}{n_S}

    where :math:`n_S` is the set of community nodes.

    :param graph: a networkx/igraph object
    :param community: NodeClustering object
    :param summary: boolean. If **True** it is returned an aggregated score for the partition is returned, otherwise individual-community ones. Default **True**.
    :return: If **summary==True** a FitnessResult object, otherwise a list of floats.

    Example:

    >>> from cdlib.algorithms import louvain
    >>> from cdlib import evaluation
    >>> g = nx.karate_club_graph()
    >>> communities = louvain(g)
    >>> mod = evaluation.triangle_participation_ratio(g,communities)

    :References:

    1. Yang, J., Leskovec, J.: Defining and evaluating network communities based on ground-truth. Knowledge and Information Systems 42(1), 181–213 (2015)
    """

    graph = convert_graph_formats(graph, nx.Graph)
    values = []
    for com in community.communities:
        coms = nx.subgraph(graph, com)
        cls = nx.triangles(coms)
        nc = [n for n in cls if cls[n] > 0]
        score = float(len(nc)) / len(coms)
        values.append(score)

    if summary:
        return FitnessResult(
            min=min(values), max=max(values), score=np.mean(values), std=np.std(values)
        )
    return values


def link_modularity(graph: nx.Graph, communities: object, **kwargs: dict) -> object:
    """
    Quality function designed for directed graphs with overlapping communities.

    :param graph: a networkx/igraph object
    :param communities: NodeClustering object
    :return: FitnessResult object

    Example:

    >>> from cdlib.algorithms import louvain
    >>> from cdlib import evaluation
    >>> g = nx.karate_club_graph()
    >>> communities = louvain(g)
    >>> mod = evaluation.link_modularity(g,communities)

    :References:

    1. Nicosia, V., Mangioni, G., Carchiolo, V., Malgeri, M.: Extending the definition of modularity to directed graphs with overlapping communities. Journal of Statistical Mechanics: Theory and Experiment 2009(03), 03024 (2009)

    """

    graph = convert_graph_formats(graph, nx.Graph)

    return FitnessResult(score=cal_modularity(graph, communities.communities))


def newman_girvan_modularity(
    graph: nx.Graph, communities: object, **kwargs: dict
) -> object:
    """Difference the fraction of intra community edges of a partition with the expected number of such edges if distributed according to a null model.

    In the standard version of modularity, the null model preserves the expected degree sequence of the graph under consideration. In other words, the modularity compares the real network structure with a corresponding one where nodes are connected without any preference about their neighbors.

    .. math:: Q(S) = \\frac{1}{m}\\sum_{c \\in S}(m_S - \\frac{(2 m_S + l_S)^2}{4m})

    where :math:`m` is the number of graph edges, :math:`m_S` is the number of community edges, :math:`l_S` is the number of edges from nodes in S to nodes outside S.

    :param graph: a networkx/igraph object
    :param communities: NodeClustering object
    :return: FitnessResult object


    Example:

    >>> from cdlib.algorithms import louvain
    >>> from cdlib import evaluation
    >>> g = nx.karate_club_graph()
    >>> communities = louvain(g)
    >>> mod = evaluation.newman_girvan_modularity(g,communities)

    :References:

    1. Newman, M.E.J. & Girvan, M. `Finding and evaluating community structure in networks. <https://www.ncbi.nlm.nih.gov/pubmed/14995526/>`_ Physical Review E 69, 26113(2004).
    """

    graph = convert_graph_formats(graph, nx.Graph)
    coms = {}
    for cid, com in enumerate(communities.communities):
        for node in com:
            coms[node] = cid

    inc = dict([])
    deg = dict([])
    links = graph.size(weight="weight")
    if links == 0:
        raise ValueError("A graph without link has an undefined modularity")

    for node in graph:
        try:
            com = coms[node]
            deg[com] = deg.get(com, 0.0) + graph.degree(node, weight="weight")
            for neighbor, dt in graph[node].items():
                weight = dt.get("weight", 1)
                if coms[neighbor] == com:
                    if neighbor == node:
                        inc[com] = inc.get(com, 0.0) + float(weight)
                    else:
                        inc[com] = inc.get(com, 0.0) + float(weight) / 2.0
        except:
            pass

    res = 0.0
    for com in set(coms.values()):
        res += (inc.get(com, 0.0) / links) - (deg.get(com, 0.0) / (2.0 * links)) ** 2

    return FitnessResult(score=res)


def erdos_renyi_modularity(
    graph: nx.Graph, communities: object, **kwargs: dict
) -> object:
    """Erdos-Renyi modularity is a variation of the Newman-Girvan one.
    It assumes that vertices in a network are connected randomly with a constant probability :math:`p`.

    .. math:: Q(S) = \\frac{1}{m}\\sum_{c \\in S} (m_S − \\frac{mn_S(n_S −1)}{n(n−1)})

    where :math:`m` is the number of graph edges, :math:`m_S` is the number of community edges, :math:`l_S` is the number of edges from nodes in S to nodes outside S.

    :param graph: a networkx/igraph object
    :param communities: NodeClustering object
    :return: FitnessResult object

    Example:

    >>> from cdlib.algorithms import louvain
    >>> from cdlib import evaluation
    >>> g = nx.karate_club_graph()
    >>> communities = louvain(g)
    >>> mod = evaluation.erdos_renyi_modularity(g,communities)

    :References:

    1. Erdos, P., & Renyi, A. (1959). `On random graphs I. <https://gnunet.org/sites/default/files/Erd%C5%91s%20%26%20R%C3%A9nyi%20-%20On%20Random%20Graphs.pdf/>`_ Publ. Math. Debrecen, 6, 290-297.
    """
    graph = convert_graph_formats(graph, nx.Graph)
    m = graph.number_of_edges()
    n = graph.number_of_nodes()
    q = 0

    for community in communities.communities:
        c = nx.subgraph(graph, community)
        mc = c.number_of_edges()
        nc = c.number_of_nodes()
        q += mc - (m * nc * (nc - 1)) / (n * (n - 1))

    return FitnessResult(score=(1 / m) * q)


def modularity_density(
    graph: nx.Graph, communities: object, lmbd: float = 0.5, **kwargs: dict
) -> object:
    """The modularity density is one of several propositions that envisioned to palliate the resolution limit issue of modularity based measures.
    The idea of this metric is to include the information about community size into the expected density of community to avoid the negligence of small and dense communities.
    For each community :math:`C` in partition :math:`S`, it uses the average modularity degree calculated by :math:`d(C) = d^{int(C)} − d^{ext(C)}` where :math:`d^{int(C)}` and :math:`d^{ext(C)}` are the average internal and external degrees of :math:`C` respectively to evaluate the fitness of :math:`C` in its network.
    Finally, the modularity density can be calculated as follows:

    .. math:: Q(S) = \\sum_{C \\in S} \\frac{1}{n_C} ( \\sum_{i \\in C} 2 * \lambda * k^{int}_{iC} - \\sum_{i \\in C} 2 * (1 - \lambda) * k^{out}_{iC})

    where :math:`n_C` is the number of nodes in C, :math:`k^{int}_{iC}` is the degree of node i within :math:`C`, :math:`k^{out}_{iC}` is the deree of node i outside :math:`C` and :math:`\lambda` is a paramter that allows for tuning the measure resolution (its default value, 0.5, computes the standard modularity density score).

    :param graph: a networkx/igraph object
    :param communities: NodeClustering object
    :param lmbd: resolution parameter, float in [0,1]. Default 0.5.
    :return: FitnessResult object


    Example:

    >>> from cdlib.algorithms import louvain
    >>> from cdlib import evaluation
    >>> g = nx.karate_club_graph()
    >>> communities = louvain(g)
    >>> mod = evaluation.modularity_density(g,communities)

    :References:

    1. Zhang, S., Ning, XM., Ding, C. et al. Determining modular organization of protein interaction networks by maximizing modularity density. <https://doi.org/10.1186/1752-0509-4-S2-S10>`_ BMC Syst Biol 4, S10 (2010).
    """
    graph = convert_graph_formats(graph, nx.Graph)
    q = 0

    for community in communities.communities:
        c = nx.subgraph(graph, community)

        nc = c.number_of_nodes()
        dint = []
        dext = []
        for node in c:
            dint.append(c.degree(node))
            dext.append(graph.degree(node) - c.degree(node))

        try:
            q += (1 / nc) * (
                (2 * lmbd * np.sum(dint)) - (2 * (1 - lmbd) * np.sum(dext))
            )
        except ZeroDivisionError:
            pass

    return FitnessResult(score=q)


def z_modularity(graph: nx.Graph, communities: object, **kwargs: dict) -> object:
    """Z-modularity is another variant of the standard modularity proposed to avoid the resolution limit.
    The concept of this version is based on an observation that the difference between the fraction of edges inside communities and the expected number of such edges in a null model should not be considered as the only contribution to the final quality of community structure.

    :param graph: a networkx/igraph object
    :param communities: NodeClustering object
    :return: FitnessResult object

    Example:

    >>> from cdlib.algorithms import louvain
    >>> from cdlib import evaluation
    >>> g = nx.karate_club_graph()
    >>> communities = louvain(g)
    >>> mod = evaluation.z_modularity(g,communities)


    :References:

    1. Miyauchi, Atsushi, and Yasushi Kawase. `Z-score-based modularity for community detection in networks. <https://journals.plos.org/plosone/article?id=10.1371/journal.pone.0147805/>`_ PloS one 11.1 (2016): e0147805.
    """
    graph = convert_graph_formats(graph, nx.Graph)
    m = graph.number_of_edges()

    mmc = 0
    dc2m = 0

    for community in communities.communities:
        c = nx.subgraph(graph, community)
        mc = c.number_of_edges()
        dc = 0

        for node in c:
            dc += graph.degree(node)

        mmc += mc / m
        dc2m += (dc / (2 * m)) ** 2

    res = 0
    try:
        res = (mmc - dc2m) / np.sqrt(dc2m * (1 - dc2m))
    except ZeroDivisionError:
        pass

    return FitnessResult(score=res)


def surprise(graph: nx.Graph, communities: object, **kwargs: dict) -> object:
    """Surprise is statistical approach proposes a quality metric assuming that edges between vertices emerge randomly according to a hyper-geometric distribution.

    According to the Surprise metric, the higher the score of a partition, the less likely it is resulted from a random realization, the better the quality of the community structure.

    :param graph: a networkx/igraph object
    :param communities: NodeClustering object
    :return: FitnessResult object

    Example:

    >>> from cdlib.algorithms import louvain
    >>> from cdlib import evaluation
    >>> g = nx.karate_club_graph()
    >>> communities = louvain(g)
    >>> mod = evaluation.surprise(g,communities)

    :References:

    1. Traag, V. A., Aldecoa, R., & Delvenne, J. C. (2015). `Detecting communities using asymptotical surprise. <https://link.aps.org/doi/10.1103/PhysRevE.92.022816/>`_ Physical Review E, 92(2), 022816.
    """
    graph = convert_graph_formats(graph, nx.Graph)
    m = graph.number_of_edges()
    n = graph.number_of_nodes()

    q = 0
    qa = 0
    sp = 0

    for community in communities.communities:
        c = nx.subgraph(graph, community)
        mc = c.number_of_edges()
        nc = c.number_of_nodes()

        q += mc
        qa += scipy.special.comb(nc, 2, exact=True)
    try:
        q = q / m
        qa = qa / scipy.special.comb(n, 2, exact=True)

        sp = m * (q * np.log(q / qa) + (1 - q) * np.log((1 - q) / (1 - qa)))
    except ZeroDivisionError:
        pass

    return FitnessResult(score=sp)


def significance(graph: nx.Graph, communities: object, **kwargs: dict) -> object:
    """Significance estimates how likely a partition of dense communities appear in a random graph.

    :param graph: a networkx/igraph object
    :param communities: NodeClustering object
    :return: FitnessResult object

    Example:

    >>> from cdlib.algorithms import louvain
    >>> from cdlib import evaluation
    >>> g = nx.karate_club_graph()
    >>> communities = louvain(g)
    >>> mod = evaluation.significance(g,communities)

    :References:

    1. Traag, V. A., Aldecoa, R., & Delvenne, J. C. (2015). `Detecting communities using asymptotical surprise. <https://link.aps.org/doi/10.1103/PhysRevE.92.022816/>`_ Physical Review E, 92(2), 022816.
    """
    graph = convert_graph_formats(graph, nx.Graph)
    m = graph.number_of_edges()

    binom = scipy.special.comb(m, 2, exact=True)
    p = m / binom

    q = 0

    for community in communities.communities:
        try:
            c = nx.subgraph(graph, community)
            nc = c.number_of_nodes()
            mc = c.number_of_edges()

            binom_c = scipy.special.comb(nc, 2, exact=True)
            pc = mc / binom_c

            q += binom_c * (pc * np.log(pc / p) + (1 - pc) * np.log((1 - pc) / (1 - p)))
        except ZeroDivisionError:
            pass
    return FitnessResult(score=q)


def purity(communities: object) -> FitnessResult:
    """Purity is the product of the frequencies of the most frequent labels carried by the nodes within the communities

    :param communities: AttrNodeClustering object
    :return: FitnessResult object

    Example:

    >>> from cdlib.algorithms import eva
    >>> from cdlib import evaluation
    >>> import random
    >>> l1 = ['A', 'B', 'C', 'D']
    >>> l2 = ["E", "F", "G"]
    >>> g = nx.barabasi_albert_graph(100, 5)
    >>> labels=dict()
    >>> for node in g.nodes():
    >>>    labels[node]={"l1":random.choice(l1), "l2":random.choice(l2)}
    >>> communities = eva(g_attr, labels, alpha=0.5)
    >>> pur = evaluation.purity(communities)

    :References:

    1. Citraro, Salvatore, and Giulio Rossetti. "Eva: Attribute-Aware Network Segmentation." International Conference on Complex Networks and Their Applications. Springer, Cham, 2019.
    """

    pur = Eva.purity(communities.coms_labels)
    return FitnessResult(score=pur)


def modularity_overlap(
    graph: nx.Graph, communities: object, weight: str = None
) -> FitnessResult:
    """Determines the Overlapping Modularity of a partition C on a graph G.

    Overlapping Modularity is defined as

     .. math:: M_{c_{r}}^{ov} = \\sum_{i \\in c_{r}} \\frac{\\sum_{j \\in c_{r}, i \\neq j}a_{ij} - \\sum_{j \\not \\in c_{r}}a_{ij}}{d_{i} \\cdot s_{i}} \\cdot \\frac{n_{c_{r}}^{e}}{n_{c_{r}} \\cdot \\binom{n_{c_{r}}}{2}}

    :param graph: a networkx/igraph object
    :param communities: NodeClustering object
    :param weight: label identifying the edge weight parameter name (if present), default None
    :return: FitnessResult object


    Example:

    >>> from cdlib.algorithms import louvain
    >>> from cdlib import evaluation
    >>> g = nx.karate_club_graph()
    >>> communities = louvain(g)
    >>> mod = evaluation.modularity_overlap(g, communities)

    :References:

    1. A. Lazar, D. Abel and T. Vicsek, "Modularity measure of networks with overlapping communities"  EPL, 90 (2010) 18001 doi: 10.1209/0295-5075/90/18001

    .. note:: Reference implementation: https://github.com/aonghus/nxtools/blob/master/nxtools/algorithms/community/quality.py
    """

    graph = convert_graph_formats(graph, nx.Graph)

    affiliation_dict = defaultdict(list)
    for cid, coms in enumerate(communities.communities):
        for n in coms:
            affiliation_dict[n].append(cid)

    mOvTotal = 0

    for nodes in communities.communities:
        nCommNodes = len(nodes)

        # the contribution of communities with 1 node is 0
        if nCommNodes <= 1:
            continue

        nInwardEdges = 0
        commStrength = 0

        for node in nodes:
            degree, inwardEdges, outwardEdges = 0, 0, 0
            for u, v, data in graph.edges(node, data=True):
                w = data.get(weight, 1)
                degree += w
                if v in nodes:
                    inwardEdges += w
                    nInwardEdges += 1
                else:
                    outwardEdges += w

            affiliationCount = len(affiliation_dict[node])
            denom = degree * affiliationCount
            if denom > 0:
                commStrength += (inwardEdges - outwardEdges) / denom

        binomC = nCommNodes * (nCommNodes - 1)
        v1 = commStrength / nCommNodes
        v2 = nInwardEdges / binomC
        mOv = v1 * v2
        mOvTotal += mOv

    score = mOvTotal / len(communities.communities)
    return FitnessResult(score=score)



from cdlib.classes.clustering import Clustering
from cdlib import evaluation
import networkx as nx

try:
    import igraph as ig
except ModuleNotFoundError:
    ig = None

from collections import defaultdict


class NodeClustering(Clustering):
    """Node Communities representation.

    :param communities: list of communities
    :param graph: a networkx/igraph object
    :param method_name: community discovery algorithm name
    :param method_parameters: configuration for the community discovery algorithm used
    :param overlap: boolean, whether the partition is overlapping or not
    """

    def __init__(
        self,
        communities: list,
        graph: object,
        method_name: str = "",
        method_parameters: dict = None,
        overlap: bool = False,
    ):
        super().__init__(communities, graph, method_name, method_parameters, overlap)

        if graph is not None:
            node_count = len(
                {nid: None for community in communities for nid in community}
            )
            if isinstance(graph, nx.Graph):
                self.node_coverage = node_count / graph.number_of_nodes()
            elif ig is not None and isinstance(graph, ig.Graph):
                self.node_coverage = node_count / graph.vcount()
            else:
                raise ValueError("Unsupported Graph type.")

    def __check_graph(self) -> bool:
        return self.graph is not None

    def to_node_community_map(self) -> dict:
        """
        Generate a <node, list(communities)> representation of the current clustering

        :return: dict of the form <node, list(communities)>
        """

        node_to_communities = defaultdict(list)
        for cid, community in enumerate(self.communities):
            for node in community:
                node_to_communities[node].append(cid)

        return node_to_communities

    def link_modularity(self) -> evaluation.FitnessResult:
        """
        Quality function designed for directed graphs with overlapping communities.

        :return: the link modularity score

        :Example:

        >>> from cdlib import evaluation
        >>> from cdlib.algorithms import louvain
        >>> g = nx.karate_club_graph()
        >>> communities = louvain(g)
        >>> mod = communities.link_modularity()

        """
        if self.__check_graph():
            return evaluation.link_modularity(self.graph, self)
        else:
            raise ValueError("Graph instance not specified")

    def normalized_cut(self, **kwargs: dict) -> evaluation.FitnessResult:
        """
        Normalized variant of the Cut-Ratio

        .. math:: : f(S) = \\frac{c_S}{2m_S+c_S} + \\frac{c_S}{2(m−m_S )+c_S}

        where :math:`m` is the number of graph edges, :math:`m_S` is the number of algorithms internal edges and :math:`c_S` is the number of algorithms nodes.

        :param summary: (optional, default True) if **True**, an overall summary is returned for the partition (min, max, avg, std); if **False** a list of community-wise score
        :return: a FitnessResult object/a list of community-wise score

        :Example:

        >>> from cdlib.algorithms import louvain
        >>> g = nx.karate_club_graph()
        >>> communities = louvain(g)
        >>> mod = communities.normalized_cut()

        """
        if self.__check_graph():
            return evaluation.normalized_cut(self.graph, self, **kwargs)
        else:
            raise ValueError("Graph instance not specified")

    def internal_edge_density(self, **kwargs: dict) -> evaluation.FitnessResult:
        """
        The internal density of the algorithms set.

        .. math:: f(S) = \\frac{m_S}{n_S(n_S−1)/2}

        where :math:`m_S` is the number of algorithms internal edges and :math:`n_S` is the number of algorithms nodes.

        :param summary: (optional, default True) if **True**, an overall summary is returned for the partition (min, max, avg, std); if **False** a list of community-wise score
        :return: a FitnessResult object/a list of community-wise score

        :Example:

        >>> from cdlib.algorithms import louvain
        >>> g = nx.karate_club_graph()
        >>> communities = louvain(g)
        >>> mod = communities.internal_edge_density()

        """
        if self.__check_graph():
            return evaluation.internal_edge_density(self.graph, self, **kwargs)
        else:
            raise ValueError("Graph instance not specified")

    def average_internal_degree(self, **kwargs: dict) -> evaluation.FitnessResult:
        """
        The average internal degree of the algorithms set.

        .. math:: f(S) = \\frac{2m_S}{n_S}

        where :math:`m_S` is the number of algorithms internal edges and :math:`n_S` is the number of algorithms nodes.

        :param summary: (optional, default True) if **True**, an overall summary is returned for the partition (min, max, avg, std); if **False** a list of community-wise score
        :return: a FitnessResult object/a list of community-wise score

        :Example:

        >>> from cdlib.algorithms import louvain
        >>> g = nx.karate_club_graph()
        >>> communities = louvain(g)
        >>> mod = communities.average_internal_degree()


        """
        if self.__check_graph():
            return evaluation.average_internal_degree(self.graph, self, **kwargs)
        else:
            raise ValueError("Graph instance not specified")

    def fraction_over_median_degree(self, **kwargs: dict) -> evaluation.FitnessResult:
        """
        Fraction of algorithms nodes of having internal degree higher than the median degree value.

        .. math:: f(S) = \\frac{|\{u: u \\in S,| \{(u,v): v \\in S\}| > d_m\}| }{n_S}


        where :math:`d_m` is the internal degree median value

        :param summary: (optional, default True) if **True**, an overall summary is returned for the partition (min, max, avg, std); if **False** a list of community-wise score
        :return: a FitnessResult object/a list of community-wise score

        :Example:

        >>> from cdlib.algorithms import louvain
        >>> g = nx.karate_club_graph()
        >>> communities = louvain(g)
        >>> mod = communities.fraction_over_median_degree()

        """
        if self.__check_graph():
            return evaluation.fraction_over_median_degree(self.graph, self, **kwargs)
        else:
            raise ValueError("Graph instance not specified")

    def expansion(self, **kwargs: dict) -> evaluation.FitnessResult:
        """
        Number of edges per algorithms node that point outside the cluster.

        .. math:: f(S) = \\frac{c_S}{n_S}

        where :math:`n_S` is the number of edges on the algorithms boundary, :math:`c_S` is the number of algorithms nodes.

        :param summary: (optional, default True) if **True**, an overall summary is returned for the partition (min, max, avg, std); if **False** a list of community-wise score
        :return: a FitnessResult object/a list of community-wise score

        :Example:

        >>> from cdlib.algorithms import louvain
        >>> g = nx.karate_club_graph()
        >>> communities = louvain(g)
        >>> mod = communities.expansion()

        """
        if self.__check_graph():
            return evaluation.expansion(self.graph, self, **kwargs)
        else:
            raise ValueError("Graph instance not specified")

    def cut_ratio(self, **kwargs: dict) -> evaluation.FitnessResult:
        """
        Fraction of existing edges (out of all possible edges) leaving the algorithms.

        ..math:: f(S) = \\frac{c_S}{n_S (n − n_S)}

        where :math:`c_S` is the number of algorithms nodes and, :math:`n_S` is the number of edges on the algorithms boundary

        :param summary: (optional, default True) if **True**, an overall summary is returned for the partition (min, max, avg, std); if **False** a list of community-wise score
        :return: a FitnessResult object/a list of community-wise score

        :Example:

        >>> from cdlib.algorithms import louvain
        >>> g = nx.karate_club_graph()
        >>> communities = louvain(g)
        >>> mod = communities.cut_ratio()

        """
        if self.__check_graph():
            return evaluation.cut_ratio(self.graph, self, **kwargs)
        else:
            raise ValueError("Graph instance not specified")

    def edges_inside(self, **kwargs: dict) -> evaluation.FitnessResult:
        """
        Number of edges internal to the algorithms.

        :param summary: (optional, default True) if **True**, an overall summary is returned for the partition (min, max, avg, std); if **False** a list of community-wise score
        :return: a FitnessResult object/a list of community-wise score

        :Example:

        >>> from cdlib.algorithms import louvain
        >>> g = nx.karate_club_graph()
        >>> communities = louvain(g)
        >>> mod = communities.edges_inside()

        """
        if self.__check_graph():
            return evaluation.edges_inside(self.graph, self, **kwargs)
        else:
            raise ValueError("Graph instance not specified")

    def conductance(self, **kwargs: dict) -> evaluation.FitnessResult:
        """
        Fraction of total edge volume that points outside the algorithms.

        .. math:: f(S) = \\frac{c_S}{2 m_S+c_S}

        where :math:`c_S` is the number of algorithms nodes and, :math:`m_S` is the number of algorithms edges

        :param summary: (optional, default True) if **True**, an overall summary is returned for the partition (min, max, avg, std); if **False** a list of community-wise score
        :return: a FitnessResult object/a list of community-wise score

        :Example:

        >>> from cdlib.algorithms import louvain
        >>> g = nx.karate_club_graph()
        >>> communities = louvain(g)
        >>> mod = communities.conductance()

        """
        if self.__check_graph():
            return evaluation.conductance(self.graph, self, **kwargs)
        else:
            raise ValueError("Graph instance not specified")

    def max_odf(self, **kwargs: dict) -> evaluation.FitnessResult:
        """
        Maximum fraction of edges of a node of a algorithms that point outside the algorithms itself.

        .. math:: max_{u \\in S} \\frac{|\{(u,v)\\in E: v \\not\\in S\}|}{d(u)}

        where :math:`E` is the graph edge set, :math:`v` is a node in :math:`S` and :math:`d(u)` is the degree of :math:`u`

        :param summary: (optional, default True) if **True**, an overall summary is returned for the partition (min, max, avg, std); if **False** a list of community-wise score
        :return: a FitnessResult object/a list of community-wise score

        :Example:

        >>> from cdlib.algorithms import louvain
        >>> g = nx.karate_club_graph()
        >>> communities = louvain(g)
        >>> mod = communities.max_odf()

        """
        if self.__check_graph():
            return evaluation.max_odf(self.graph, self, **kwargs)
        else:
            raise ValueError("Graph instance not specified")

    def avg_odf(self, **kwargs: dict) -> evaluation.FitnessResult:
        """
        Average fraction of edges of a node of a algorithms that point outside the algorithms itself.

        .. math:: \\frac{1}{n_S} \\sum_{u \\in S} \\frac{|\{(u,v)\\in E: v \\not\\in S\}|}{d(u)}

        where :math:`E` is the graph edge set, :math:`v` is a node in :math:`S`, :math:`d(u)` is the degree of :math:`u` and :math:`n_S` is the set of algorithms nodes.

        :param summary: (optional, default True) if **True**, an overall summary is returned for the partition (min, max, avg, std); if **False** a list of community-wise score
        :return: a FitnessResult object/a list of community-wise score
        :Example:

        >>> from cdlib.algorithms import louvain
        >>> g = nx.karate_club_graph()
        >>> communities = louvain(g)
        >>> mod = communities.avg_odf()

        """
        if self.__check_graph():
            return evaluation.avg_odf(self.graph, self, **kwargs)
        else:
            raise ValueError("Graph instance not specified")

    def flake_odf(self, **kwargs: dict) -> evaluation.FitnessResult:
        """
        Fraction of nodes in S that have fewer edges pointing inside than to the outside of the algorithms.

        .. math:: f(S) = \\frac{| \{ u:u \in S,| \{(u,v) \in E: v \in S \}| < d(u)/2 \}|}{n_S}

        where :math:`E` is the graph edge set, :math:`v` is a node in :math:`S`, :math:`d(u)` is the degree of :math:`u` and :math:`n_S` is the set of algorithms nodes.

        :param summary: (optional, default True) if **True**, an overall summary is returned for the partition (min, max, avg, std); if **False** a list of community-wise score
        :return: a FitnessResult object/a list of community-wise score

        :Example:

        >>> from cdlib.algorithms import louvain
        >>> g = nx.karate_club_graph()
        >>> communities = louvain(g)
        >>> mod = communities.flake_odf()

        """
        if self.__check_graph():
            return evaluation.flake_odf(self.graph, self, **kwargs)
        else:
            raise ValueError("Graph instance not specified")

    def triangle_participation_ratio(self, **kwargs: dict) -> evaluation.FitnessResult:
        """
        Fraction of algorithms nodes that belong to a triad.

        .. math:: f(S) = \\frac{ | \{ u: u \in S,\{(v,w):v, w \in S,(u,v) \in E,(u,w) \in E,(v,w) \in E \} \\not = \\emptyset \} |}{n_S}

        where :math:`n_S` is the set of algorithms nodes.

        :param summary: (optional, default True) if **True**, an overall summary is returned for the partition (min, max, avg, std); if **False** a list of community-wise score
        :return: a FitnessResult object/a list of community-wise score

        :Example:

        >>> from cdlib.algorithms import louvain
        >>> g = nx.karate_club_graph()
        >>> communities = louvain(g)
        >>> mod = communities.triangle_participation_ratio()

        """
        if self.__check_graph():
            return evaluation.triangle_participation_ratio(self.graph, self, **kwargs)
        else:
            raise ValueError("Graph instance not specified")

    def size(self, **kwargs: dict) -> evaluation.FitnessResult:
        """Size is the number of nodes in the community

        :param summary: (optional, default True) if **True**, an overall summary is returned for the partition (min, max, avg, std); if **False** a list of community-wise score
        :return: a FitnessResult object/a list of community-wise score

        Example:

        >>> from cdlib.algorithms import louvain
        >>> g = nx.karate_club_graph()
        >>> communities = louvain(g)
        >>> mod = communities.size()
        """

        return evaluation.size(self.graph, self, **kwargs)

    def newman_girvan_modularity(self) -> evaluation.FitnessResult:
        """
        Difference the fraction of intra algorithms edges of a partition with the expected number of such edges if distributed according to a null model.

        In the standard version of modularity, the null model preserves the expected degree sequence of the graph under consideration. In other words, the modularity compares the real network structure with a corresponding one where nodes are connected without any preference about their neighbors.

        .. math:: Q(S) = \\frac{1}{m}\\sum_{c \\in S}(m_S - \\frac{(2 m_S + l_S)^2}{4m})

        where :math:`m` is the number of graph edges, :math:`m_S` is the number of algorithms edges, :math:`l_S` is the number of edges from nodes in S to nodes outside S.


        :return: the Newman-Girvan modularity score

        :Example:

        >>> from cdlib.algorithms import louvain
        >>> g = nx.karate_club_graph()
        >>> communities = louvain(g)
        >>> mod = communities.newman_girvan_modularity()

        :References:

        Newman, M.E.J. & Girvan, M. **Finding and evaluating algorithms structure in networks.** Physical Review E 69, 26113(2004).
        """
        if self.__check_graph():
            return evaluation.newman_girvan_modularity(self.graph, self)
        else:
            raise ValueError("Graph instance not specified")

    def erdos_renyi_modularity(self) -> evaluation.FitnessResult:
        """

        Erdos-Renyi modularity is a variation of the Newman-Girvan one.
        It assumes that vertices in a network are connected randomly with a constant probability :math:`p`.

        .. math:: Q(S) = \\frac{1}{m}\\sum_{c \\in S} (m_S − \\frac{mn_S(n_S −1)}{n(n−1)})

        where :math:`m` is the number of graph edges, :math:`m_S` is the number of algorithms edges, :math:`l_S` is the number of edges from nodes in S to nodes outside S.


        :return: the Erdos-Renyi modularity score

        :Example:

        >>> from cdlib.algorithms import louvain
        >>> g = nx.karate_club_graph()
        >>> communities = louvain(g)
        >>> mod = communities.erdos_renyi_modularity()

        :References:

        Erdos, P., & Renyi, A. (1959). **On random graphs I.** Publ. Math. Debrecen, 6, 290-297.
        """
        if self.__check_graph():
            return evaluation.erdos_renyi_modularity(self.graph, self)
        else:
            raise ValueError("Graph instance not specified")

    def modularity_density(self) -> evaluation.FitnessResult:
        """
        The modularity density is one of several propositions that envisioned to palliate the resolution limit issue of modularity based measures.
        The idea of this metric is to include the information about algorithms size into the expected density of algorithms to avoid the negligence of small and dense communities.
        For each algorithms :math:`C` in partition :math:`S`, it uses the average modularity degree calculated by :math:`d(C) = d^{int(C)} − d^{ext(C)}` where :math:`d^{int(C)}` and :math:`d^{ext(C)}` are the average internal and external degrees of :math:`C` respectively to evaluate the fitness of :math:`C` in its network.
        Finally, the modularity density can be calculated as follows:

        .. math:: Q(S) = \\sum_{C \\in S} \\frac{1}{n_C} ( \\sum_{i \\in C} k^{int}_{iC} - \\sum_{i \\in C} k^{out}_{iC})

        where :math:`n_C` is the number of nodes in C, :math:`k^{int}_{iC}` is the degree of node i within :math:`C` and :math:`k^{out}_{iC}` is the deree of node i outside :math:`C`.


        :return: the modularity density score

        :Example:

        >>> from cdlib.algorithms import louvain
        >>> g = nx.karate_club_graph()
        >>> communities = louvain(g)
        >>> mod = communities.modularity_density()

        :References:

        Li, Z., Zhang, S., Wang, R. S., Zhang, X. S., & Chen, L. (2008). **Quantitative function for algorithms detection.** Physical review E, 77(3), 036109.

        """

        if self.__check_graph():
            return evaluation.modularity_density(self.graph, self)
        else:
            raise ValueError("Graph instance not specified")

    def z_modularity(self) -> evaluation.FitnessResult:
        """
        Z-modularity is another variant of the standard modularity proposed to avoid the resolution limit.
        The concept of this version is based on an observation that the difference between the fraction of edges inside communities and the expected number of such edges in a null model should not be considered as the only contribution to the final quality of algorithms structure.

        :return: the z-modularity score

        :Example:

        >>> from cdlib.algorithms import louvain
        >>> g = nx.karate_club_graph()
        >>> communities = louvain(g)
        >>> mod = communities.z_modularity()


        :References:

        Miyauchi, Atsushi, and Yasushi Kawase. **Z-score-based modularity for algorithms detection in networks.** PloS one 11.1 (2016): e0147805.
        """

        if self.__check_graph():
            return evaluation.z_modularity(self.graph, self)
        else:
            raise ValueError("Graph instance not specified")

    def modularity_overlap(self, weight: str = None) -> evaluation.FitnessResult:
        """Determines the Overlapping Modularity of a partition C on a graph G.

        Overlapping Modularity is defined as

        .. math:: M_{c_{r}}^{ov} = \\sum_{i \\in c_{r}} \\frac{\\sum_{j \\in c_{r}, i \\neq j}a_{ij} - \\sum_{j \\not \\in c_{r}}a_{ij}}{d_{i} \\cdot s_{i}} \\cdot \\frac{n_{c_{r}}^{e}}{n_{c_{r}} \\cdot \\binom{n_{c_{r}}}{2}}

        :param weight: label identifying the edge weight parameter name (if present), default None
        :return: FitnessResult object

        Example:

        >>> from cdlib.algorithms import louvain
        >>> from cdlib import evaluation
        >>> g = nx.karate_club_graph()
        >>> communities = louvain(g)
        >>> mod = communities.modularity_overlap()

        :References:

        1. A. Lazar, D. Abel and T. Vicsek, "Modularity measure of networks with overlapping communities"  EPL, 90 (2010) 18001 doi: 10.1209/0295-5075/90/18001
        """

        if self.__check_graph():
            return evaluation.modularity_overlap(self.graph, self, weight)
        else:
            raise ValueError("Graph instance not specified")

    def surprise(self) -> evaluation.FitnessResult:
        """
        Surprise is statistical approach proposes a quality metric assuming that edges between vertices emerge randomly according to a hyper-geometric distribution.

        According to the Surprise metric, the higher the score of a partition, the less likely it is resulted from a random realization, the better the quality of the algorithms structure.

        :return: the surprise score

        :Example:

        >>> from cdlib.algorithms import louvain
        >>> g = nx.karate_club_graph()
        >>> communities = louvain(g)
        >>> mod = communities.surprise()

        :References:

        Traag, V. A., Aldecoa, R., & Delvenne, J. C. (2015). **Detecting communities using asymptotical surprise.** Physical Review E, 92(2), 022816.
        """

        if self.__check_graph():
            return evaluation.surprise(self.graph, self)
        else:
            raise ValueError("Graph instance not specified")

    def significance(self) -> evaluation.FitnessResult:
        """
        Significance estimates how likely a partition of dense communities appear in a random graph.

        :return: the significance score

        :Example:

        >>> from cdlib.algorithms import louvain
        >>> g = nx.karate_club_graph()
        >>> communities = louvain(g)
        >>> mod = communities.significance()


        :References:

        Traag, V. A., Aldecoa, R., & Delvenne, J. C. (2015). **Detecting communities using asymptotical surprise.** Physical Review E, 92(2), 022816.
        """

        if self.__check_graph():
            return evaluation.significance(self.graph, self)
        else:
            raise ValueError("Graph instance not specified")

    def scaled_density(self, **kwargs: dict) -> evaluation.FitnessResult:
        """Scaled density.

        The scaled density of a community is defined as the ratio of the community density w.r.t. the complete graph density.

        :param summary: boolean. If **True** it is returned an aggregated score for the partition is returned, otherwise individual-community ones. Default **True**.
        :return: If **summary==True** a FitnessResult object, otherwise a list of floats.

        Example:

        >>> from cdlib.algorithms import louvain
        >>> from cdlib import evaluation
        >>> g = nx.karate_club_graph()
        >>> communities = louvain(g)
        >>> scd = communities.scaled_density()
        """

        if self.__check_graph():
            return evaluation.scaled_density(self.graph, self, **kwargs)
        else:
            raise ValueError("Graph instance not specified")

    def avg_distance(self, **kwargs: dict) -> evaluation.FitnessResult:
        """Average distance.

        The average distance of a community is defined average path length across all possible pair of nodes composing it.

        :param summary: boolean. If **True** it is returned an aggregated score for the partition is returned, otherwise individual-community ones. Default **True**.
        :return: If **summary==True** a FitnessResult object, otherwise a list of floats.

        Example:

        >>> from cdlib.algorithms import louvain
        >>> from cdlib import evaluation
        >>> g = nx.karate_club_graph()
        >>> communities = louvain(g)
        >>> scd = communities.avg_distance()
        """

        if self.__check_graph():
            return evaluation.avg_distance(self.graph, self, **kwargs)
        else:
            raise ValueError("Graph instance not specified")

    def hub_dominance(self, **kwargs: dict) -> evaluation.FitnessResult:
        """Hub dominance.

        The hub dominance of a community is defined as the ratio of the degree of its most connected node w.r.t. the theoretically maximal degree within the community.

        :param summary: boolean. If **True** it is returned an aggregated score for the partition is returned, otherwise individual-community ones. Default **True**.
        :return: If **summary==True** a FitnessResult object, otherwise a list of floats.

        Example:

        >>> from cdlib.algorithms import louvain
        >>> from cdlib import evaluation
        >>> g = nx.karate_club_graph()
        >>> communities = louvain(g)
        >>> scd = communities.hub_dominance()
        """

        if self.__check_graph():
            return evaluation.hub_dominance(self.graph, self, **kwargs)
        else:
            raise ValueError("Graph instance not specified")

    def avg_transitivity(self, **kwargs: dict) -> evaluation.FitnessResult:
        """Average transitivity.

        The average transitivity of a community is defined the as the average clustering coefficient of its nodes w.r.t. their connection within the community itself.

        :param summary: boolean. If **True** it is returned an aggregated score for the partition is returned, otherwise individual-community ones. Default **True**.
        :return: If **summary==True** a FitnessResult object, otherwise a list of floats.

        Example:

        >>> from cdlib.algorithms import louvain
        >>> from cdlib import evaluation
        >>> g = nx.karate_club_graph()
        >>> communities = louvain(g)
        >>> scd = communities.avg_transitivity()
        """

        if self.__check_graph():
            return evaluation.avg_transitivity(self.graph, self, **kwargs)
        else:
            raise ValueError("Graph instance not specified")

    def avg_embeddedness(self, **kwargs: dict) -> evaluation.FitnessResult:
        """Average embeddedness of nodes within the community.

        The embeddedness of a node n w.r.t. a community C is the ratio of its degree within the community and its overall degree.

        .. math:: emb(n,C) = \\frac{k_n^C}{k_n}

        The average embeddedness of a community C is:

        .. math:: avg_embd(c) = \\frac{1}{|C|} \sum_{i \in C} \\frac{k_n^C}{k_n}

        :param summary: boolean. If **True** it is returned an aggregated score for the partition is returned, otherwise individual-community ones. Default **True**.
        :return: If **summary==True** a FitnessResult object, otherwise a list of floats.

        Example:

        >>> from cdlib.algorithms import louvain
        >>> from cdlib import evaluation
        >>> g = nx.karate_club_graph()
        >>> communities = louvain(g)
        >>> ave = communities.avg_embeddedness()

        """

        if self.__check_graph():
            return evaluation.avg_embeddedness(self.graph, self, **kwargs)
        else:
            raise ValueError("Graph instance not specified")

    def normalized_mutual_information(
        self, clustering: Clustering
    ) -> evaluation.MatchingResult:
        """
        Normalized Mutual Information between two clusterings.

        Normalized Mutual Information (NMI) is an normalization of the Mutual
        Information (MI) score to scale the results between 0 (no mutual
        information) and 1 (perfect correlation). In this function, mutual
        information is normalized by ``sqrt(H(labels_true) * H(labels_pred))``

        :param clustering: NodeClustering object
        :return: normalized mutual information score

        :Example:

        >>> from cdlib.algorithms import louvain
        >>> g = nx.karate_club_graph()
        >>> communities = louvain(g)
        >>> leiden_communities = algorithms.leiden(g)
        >>> mod = communities.normalized_mutual_information(leiden_communities)

        """

        return evaluation.normalized_mutual_information(self, clustering)

    def overlapping_normalized_mutual_information_LFK(
        self, clustering: Clustering
    ) -> evaluation.MatchingResult:
        """
        Overlapping Normalized Mutual Information between two clusterings.

        Extension of the Normalized Mutual Information (NMI) score to cope with overlapping partitions.
        This is the version proposed by Lancichinetti et al.

        :param clustering: NodeClustering object
        :return: onmi score

        :Example:

        >>> from cdlib.algorithms import louvain
        >>> g = nx.karate_club_graph()
        >>> communities = louvain(g)
        >>> leiden_communities = algorithms.leiden(g)
        >>> mod = communities.overlapping_normalized_mutual_information_LFK(leiden_communities)

        :Reference:

        1. Lancichinetti, A., Fortunato, S., & Kertesz, J. (2009). Detecting the overlapping and hierarchical community structure in complex networks. New Journal of Physics, 11(3), 033015.

        """
        return evaluation.overlapping_normalized_mutual_information_LFK(
            self, clustering
        )

    def overlapping_normalized_mutual_information_MGH(
        self, clustering: Clustering, normalization: str = "max"
    ) -> evaluation.MatchingResult:
        """
        Overlapping Normalized Mutual Information between two clusterings.

        Extension of the Normalized Mutual Information (NMI) score to cope with overlapping partitions.
        This is the version proposed by McDaid et al. using a different normalization than the original LFR one. See ref.
        for more details.

        :param clustering: NodeClustering object
        :param normalization: one of "max" or "LFK". Default "max" (corresponds to the main method described in the article)
        :return: onmi score

        :Example:

        >>> from cdlib import evaluation, algorithms
        >>> g = nx.karate_club_graph()
        >>> louvain_communities = algorithms.louvain(g)
        >>> leiden_communities = algorithms.leiden(g)
        >>> evaluation.overlapping_normalized_mutual_information_MGH(louvain_communities,leiden_communities)
        :Reference:

        1. McDaid, A. F., Greene, D., & Hurley, N. (2011). Normalized mutual information to evaluate overlapping community finding algorithms. arXiv preprint arXiv:1110.2515. Chicago
        """
        return evaluation.overlapping_normalized_mutual_information_MGH(
            self, clustering, normalization=normalization
        )

    def omega(self, clustering: Clustering) -> evaluation.MatchingResult:
        """

        Index of resemblance for overlapping, complete coverage, network clusterings.

        :param clustering: NodeClustering object
        :return: omega index

        :Example:

        >>> from cdlib.algorithms import louvain
        >>> g = nx.karate_club_graph()
        >>> communities = louvain(g)
        >>> leiden_communities = algorithms.leiden(g)
        >>> mod = communities.omega(leiden_communities)


        :Reference:

        1. Gabriel Murray, Giuseppe Carenini, and Raymond Ng. 2012. **Using the omega index for evaluating abstractive algorithms detection.** In Proceedings of Workshop on Evaluation Metrics and System Comparison for Automatic Summarization. Association for Computational Linguistics, Stroudsburg, PA, USA, 10-18.
        """
        return evaluation.omega(self, clustering)

    def f1(self, clustering: Clustering) -> evaluation.MatchingResult:
        """
        Compute the average F1 score of the optimal algorithms matches among the partitions in input.
        Works on overlapping/non-overlapping complete/partial coverage partitions.

        :param clustering: NodeClustering object
        :return: F1 score (harmonic mean of precision and recall)

        :Example:

        >>> from cdlib.algorithms import louvain
        >>> g = nx.karate_club_graph()
        >>> communities = louvain(g)
        >>> leiden_communities = algorithms.leiden(g)
        >>> mod = communities.f1(leiden_communities)


        :Reference:

        1. Rossetti, G., Pappalardo, L., & Rinzivillo, S. (2016). **A novel approach to evaluate algorithms detection internal on ground truth.** In Complex Networks VII (pp. 133-144). Springer, Cham.
        """
        return evaluation.f1(self, clustering)

    def nf1(self, clustering: Clustering) -> evaluation.MatchingResult:
        """
        Compute the Normalized F1 score of the optimal algorithms matches among the partitions in input.
        Works on overlapping/non-overlapping complete/partial coverage partitions.

        :param clustering: NodeClustering object
        :return: MatchingResult instance

        :Example:

        >>> from cdlib.algorithms import louvain
        >>> g = nx.karate_club_graph()
        >>> communities = louvain(g)
        >>> leiden_communities = algorithms.leiden(g)
        >>> mod = communities.nf1(leiden_communities)


        :Reference:

        1. Rossetti, G., Pappalardo, L., & Rinzivillo, S. (2016). **A novel approach to evaluate algorithms detection internal on ground truth.**

        2. Rossetti, G. (2017). : **RDyn: graph benchmark handling algorithms dynamics. Journal of Complex Networks.** 5(6), 893-912.
        """
        return evaluation.nf1(self, clustering)

    def adjusted_rand_index(self, clustering: Clustering) -> evaluation.MatchingResult:
        """
        Rand index adjusted for chance.

        The Rand Index computes a similarity measure between two clusterings
        by considering all pairs of samples and counting pairs that are
        assigned in the same or different clusters in the predicted and
        true clusterings.

        The raw RI score is then "adjusted for chance" into the ARI score
        using the following scheme::

            ARI = (RI - Expected_RI) / (max(RI) - Expected_RI)

        The adjusted Rand index is thus ensured to have a value close to
        0.0 for random labeling independently of the number of clusters and
        samples and exactly 1.0 when the clusterings are identical (up to
        a permutation).

        ARI is a symmetric measure::

            adjusted_rand_index(a, b) == adjusted_rand_index(b, a)

        :param clustering: NodeClustering object
        :return: ARI score

        :Example:

        >>> from cdlib.algorithms import louvain
        >>> g = nx.karate_club_graph()
        >>> communities = louvain(g)
        >>> leiden_communities = algorithms.leiden(g)
        >>> mod = communities.adjusted_rand_index(leiden_communities)


        :Reference:

        1. Hubert, L., & Arabie, P. (1985). **Comparing partitions**. Journal of classification, 2(1), 193-218.
        """
        return evaluation.adjusted_rand_index(self, clustering)

    def adjusted_mutual_information(
        self, clustering: Clustering
    ) -> evaluation.MatchingResult:
        """
        Adjusted Mutual Information between two clusterings.

        Adjusted Mutual Information (AMI) is an adjustment of the Mutual
        Information (MI) score to account for chance. It accounts for the fact that
        the MI is generally higher for two clusterings with a larger number of
        clusters, regardless of whether there is actually more information shared.
        For two clusterings :math:`U` and :math:`V`, the AMI is given as::

            AMI(U, V) = [MI(U, V) - E(MI(U, V))] / [max(H(U), H(V)) - E(MI(U, V))]

        This metric is independent of the absolute values of the labels:
        a permutation of the class or cluster label values won't change the
        score value in any way.

        This metric is furthermore symmetric: switching ``label_true`` with
        ``label_pred`` will return the same score value. This can be useful to
        measure the agreement of two independent label assignments strategies
        on the same dataset when the real ground truth is not known.

        Be mindful that this function is an order of magnitude slower than other
        metrics, such as the Adjusted Rand Index.

        :param clustering: NodeClustering object
        :return: AMI score

        :Example:

        >>> from cdlib.algorithms import louvain
        >>> g = nx.karate_club_graph()
        >>> communities = louvain(g)
        >>> leiden_communities = algorithms.leiden(g)
        >>> mod = communities.adjusted_mutual_information(leiden_communities)

        :Reference:

        1. Vinh, N. X., Epps, J., & Bailey, J. (2010). **Information theoretic measures for clusterings comparison: Variants, properties, normalization and correction for chance.** Journal of Machine Learning Research, 11(Oct), 2837-2854.
        """
        return evaluation.adjusted_mutual_information(self, clustering)

    def variation_of_information(
        self, clustering: Clustering
    ) -> evaluation.MatchingResult:
        """
        Variation of Information among two nodes partitions.

        $$ H(p)+H(q)-2MI(p, q) $$

        where MI is the mutual information, H the partition entropy and p,q are the algorithms sets

        :param clustering: NodeClustering object
        :return: VI score

        :Example:

        >>> from cdlib.algorithms import louvain
        >>> g = nx.karate_club_graph()
        >>> communities = louvain(g)
        >>> leiden_communities = algorithms.leiden(g)
        >>> mod = communities.variation_of_information(leiden_communities)


        :Reference:

        1. Meila, M. (2007). **Comparing clusterings - an information based distance.** Journal of Multivariate Analysis, 98, 873-895. doi:10.1016/j.jmva.2006.11.013
        """
        return evaluation.variation_of_information(self, clustering)



from copy import copy
from heapq import heappush, heappop
from itertools import combinations, chain
from collections import defaultdict


# Hierarchical Link Community
"""
Yong-Yeol Ahn, James P. Bagrow, and Sune Lehmann, 
Link communities reveal multiscale complexity in networks, 
Nature 466, 761 (2010)
"""


def get_sorted_edge(a, b):
    return tuple(sorted([a, b]))


def HLC_read_edge_list_unweighted(g):
    adj_list_dict = defaultdict(set)
    edges = set()

    for e in g.edges():
        ni, nj = e[0], e[1]
        edges.add(get_sorted_edge(ni, nj))
        adj_list_dict[ni].add(nj)
        adj_list_dict[nj].add(ni)
    return dict(adj_list_dict), edges


def HLC_read_edge_list_weighted(g):
    adj = defaultdict(set)
    edges = set()
    ij2wij = {}
    for e in g.edges(data=True):
        ni, nj, wij = e[0], e[1], e[2]["weight"]
        if ni != nj:
            ni, nj = get_sorted_edge(ni, nj)
            edges.add((ni, nj))
            ij2wij[ni, nj] = wij
            adj[ni].add(nj)
            adj[nj].add(ni)
    return dict(adj), edges, ij2wij


class HLC(object):
    @staticmethod
    def get_sorted_pair(a, b):
        return tuple(sorted([a, b]))

    @staticmethod
    def sort_edge_pairs_by_similarity(adj_list_dict):
        def cal_jaccard(left_set, right_set):
            return 1.0 * len(left_set & right_set) / len(left_set | right_set)

        inc_adj_list_dict = dict((n, adj_list_dict[n] | {n}) for n in adj_list_dict)
        min_heap = []
        for vertex in adj_list_dict:
            if len(adj_list_dict[vertex]) > 1:
                for i, j in combinations(adj_list_dict[vertex], 2):
                    edge_pair = HLC.get_sorted_pair(
                        HLC.get_sorted_pair(i, vertex), HLC.get_sorted_pair(j, vertex)
                    )
                    similarity_ratio = cal_jaccard(
                        inc_adj_list_dict[i], inc_adj_list_dict[j]
                    )
                    heappush(min_heap, (1 - similarity_ratio, edge_pair))
        return [heappop(min_heap) for _ in range(len(min_heap))]

    def __init__(self, adj_list_dict, edges):
        self.adj_list_dict = adj_list_dict
        self.edges = edges
        self.density_factor = 2.0 / len(edges)

        self.edge2cid = {}
        self.cid2nodes, self.cid2edges = {}, {}
        self.orig_cid2edge = {}
        self.curr_max_cid = 0
        self.linkage = []  # dendrogram

        def initialize_edges():
            for cid, edge in enumerate(self.edges):
                edge = HLC.get_sorted_pair(*edge)
                self.edge2cid[edge] = cid
                self.cid2edges[cid] = {edge}
                self.orig_cid2edge[cid] = edge
                self.cid2nodes[cid] = set(edge)
            self.curr_max_cid = len(self.edges) - 1

        initialize_edges()  # every edge in its own comm

        self.D = 0.0  # partition density
        self.list_D = [(1.0, 0.0)]  # list of (S_i,D_i) tuples...
        self.best_D = 0.0
        self.best_S = 1.0  # similarity threshold at self.best_D
        self.best_P = None  # best partition, dict: edge -> cid

    def single_linkage(self, threshold=None, w=None, dendro_flag=False):
        def merge_comms(edge1, edge2, S):
            def cal_density(edge_num, vertex_num):
                return (
                    edge_num
                    * (edge_num - vertex_num + 1.0)
                    / ((vertex_num - 2.0) * (vertex_num - 1.0))
                    if vertex_num > 2
                    else 0.0
                )

            if (
                not edge1 or not edge2
            ):  # We'll get (None, None) at the end of clustering
                return
            cid1, cid2 = self.edge2cid[edge1], self.edge2cid[edge2]
            if cid1 == cid2:  # already merged!
                return
            m1, m2 = len(self.cid2edges[cid1]), len(self.cid2edges[cid2])
            n1, n2 = len(self.cid2nodes[cid1]), len(self.cid2nodes[cid2])
            Dc1, Dc2 = cal_density(m1, n1), cal_density(m2, n2)
            if m2 > m1:  # merge smaller into larger
                cid1, cid2 = cid2, cid1

            if dendro_flag:
                self.curr_max_cid += 1
                newcid = self.curr_max_cid
                self.cid2edges[newcid] = self.cid2edges[cid1] | self.cid2edges[cid2]
                self.cid2nodes[newcid] = set()
                for e in chain(self.cid2edges[cid1], self.cid2edges[cid2]):
                    self.cid2nodes[newcid] |= set(e)
                    self.edge2cid[e] = newcid
                del self.cid2edges[cid1], self.cid2nodes[cid1]
                del self.cid2edges[cid2], self.cid2nodes[cid2]
                m, n = len(self.cid2edges[newcid]), len(self.cid2nodes[newcid])

                self.linkage.append((cid1, cid2, S))

            else:
                self.cid2edges[cid1] |= self.cid2edges[cid2]
                for e in self.cid2edges[cid2]:  # move edges,nodes from cid2 to cid1
                    self.cid2nodes[cid1] |= set(e)
                    self.edge2cid[e] = cid1
                del self.cid2edges[cid2], self.cid2nodes[cid2]

                m, n = len(self.cid2edges[cid1]), len(self.cid2nodes[cid1])

            Dc12 = cal_density(m, n)
            self.D += (
                Dc12 - Dc1 - Dc2
            ) * self.density_factor  # update partition density

        if w is None:
            sorted_edge_list = HLC.sort_edge_pairs_by_similarity(self.adj_list_dict)
        else:
            sorted_edge_list = sort_edge_pairs_by_similarity_weighted(
                self.adj_list_dict, w
            )

        prev_similarity = -1
        for oms, eij_eik in chain(sorted_edge_list, [(1.0, (None, None))]):
            cur_similarity = 1 - oms
            if threshold and cur_similarity < threshold:
                break

            if cur_similarity != prev_similarity:  # update list
                if self.D >= self.best_D:  # check PREVIOUS merger, because that's
                    self.best_D = self.D  # the end of the tie
                    self.best_S = cur_similarity
                    self.best_P = copy(self.edge2cid)  # slow...
                self.list_D.append((cur_similarity, self.D))
                prev_similarity = cur_similarity
            merge_comms(eij_eik[0], eij_eik[1], cur_similarity)

        if threshold is not None:
            return self.edge2cid, self.D
        if dendro_flag:
            return (
                self.best_P,
                self.best_S,
                self.best_D,
                self.list_D,
                self.orig_cid2edge,
                self.linkage,
            )
        else:
            return self.best_P, self.best_S, self.best_D, self.list_D


def sort_edge_pairs_by_similarity_weighted(adj_list_dict, edge_weight_dict):
    inc_adj_list_dict = dict((n, adj_list_dict[n] | {n}) for n in adj_list_dict)

    def cal_jaccard(intersect_val, left_val, right_val):
        return intersect_val / (left_val + right_val - intersect_val)

    Aij = copy(edge_weight_dict)
    n2a_sqrd = {}
    for vertex in adj_list_dict:
        Aij[vertex, vertex] = float(
            sum(
                edge_weight_dict[HLC.get_sorted_pair(vertex, i)]
                for i in adj_list_dict[vertex]
            )
        )
        Aij[vertex, vertex] /= len(adj_list_dict[vertex])
        n2a_sqrd[vertex] = sum(
            Aij[HLC.get_sorted_pair(vertex, i)] ** 2 for i in inc_adj_list_dict[vertex]
        )  # includes (n,n)!

    min_heap = []
    for vertex in adj_list_dict:
        if len(adj_list_dict[vertex]) > 1:
            for i, j in combinations(adj_list_dict[vertex], 2):
                edge_pair = HLC.get_sorted_pair(
                    HLC.get_sorted_pair(i, vertex), HLC.get_sorted_pair(j, vertex)
                )
                ai_dot_aj = float(
                    sum(
                        Aij[HLC.get_sorted_pair(i, x)] * Aij[HLC.get_sorted_pair(j, x)]
                        for x in inc_adj_list_dict[i] & inc_adj_list_dict[j]
                    )
                )
                similarity_ratio = cal_jaccard(ai_dot_aj, n2a_sqrd[i], n2a_sqrd[j])
                heappush(min_heap, (1 - similarity_ratio, edge_pair))
    return [heappop(min_heap) for _ in range(len(min_heap))]



import networkx as nx
import itertools

parent = dict()
rank = dict()


def __make_set(vertice):
    parent[vertice] = vertice
    rank[vertice] = 0


def __find(vertice):
    if parent[vertice] != vertice:
        parent[vertice] = __find(parent[vertice])
    return parent[vertice]


def __union(vertice1, vertice2):
    root1 = __find(vertice1)
    root2 = __find(vertice2)
    if root1 != root2:
        if rank[root1] > rank[root2]:
            parent[root2] = root1
        else:
            parent[root1] = root2
        if rank[root1] == rank[root2]:
            rank[root2] += 1


def UMSTMO(G):
    G3 = nx.Graph()
    # list of edges in la list T
    T = list(G.edges())

    i = 0
    # joining each edge to its weight
    z2 = []
    while i < len(T):
        # z is number of common neighbors e(i,j)
        z = len(list(nx.common_neighbors(G, T[i][0], T[i][1])))
        # x the number of neighbors i and x1 for j
        x = len(list(G.neighbors(T[i][0])))
        x1 = len(list(G.neighbors(T[i][1])))
        if z > 0:
            # p is the value of jaccard coefficient
            p = z / (x + x1 + z)
            # add weight to the edge
            G[T[i][0]][T[i][1]]["weight"] = p
            # list of edges and their weights
            z2.append([p, T[i][0], T[i][1]])
        else:
            G[T[i][0]][T[i][1]]["weight"] = 0
            z2.append([0, T[i][0], T[i][1]])
        i = i + 1

    for l in G.nodes():
        __make_set(l)

    # sort the list of edges according to their weights
    z2.sort(reverse=True)
    B = []
    eu = []
    i = 0
    # construction of the union of all maximum spanning tree
    for k in z2:
        M = []
        elarge = []
        t8 = 0
        while i < len(z2):
            if z2[i][0] == k[0]:
                elarge.append([z2[i][1], z2[i][2]])
                i = i + 1
                t8 = i
            else:
                i = len(z2)
        i = t8
        for ll in elarge:
            if __find(ll[0]) != __find(ll[1]):
                M.append(ll)
        for kk in M:
            __union(kk[0], kk[1])
        eu.extend(M)
        B.append(elarge)

    tr1 = eu
    for i in tr1:
        G3.add_edge(i[0], i[1])

    l1 = []

    for i in G3.nodes():
        di = G3.degree(i)
        l1.append([di, i])

    l1.sort(reverse=True)
    ll = []
    for k in l1:
        ll.append(k[1])
    re = []
    up = []
    for k in ll:
        b = G3.neighbors(k)
        c = []
        c1 = []

        i = 0
        while i < len(list(b)):
            j = i + 1
            while j < len(list(b)):

                if G.has_edge(b[i], b[j]):
                    c1.extend([b[i], b[j]])
                j = j + 1
            i = i + 1

        c1 = list(set(c1))
        c.append(c1)
        if len(list(b)) == 1:
            if k not in up:
                c.append([b[0]])
                up.append(b[0])
        for k2 in c:
            if len(k2) > 0:
                k2.append(k)
                k2.sort()
                if k2 not in re:
                    re.append(k2)
                for ll2 in k2:
                    up.append(ll2)

    sup = list(set(G.nodes()) - set(up))
    re.sort(key=len, reverse=True)
    res = re

    for k in sup:
        max_val = 0
        ne = G.neighbors(k)
        i = 0
        temp = 0
        while i < len(res):
            aa = len(set(ne).intersection(res[i]))

            if aa >= max_val:
                max_val = aa
                temp = i
                if aa > len(list(ne)) - aa:
                    i = len(res)
            i = i + 1
        if max_val > 0:
            res[temp].append(k)
        else:
            res.append([k])

    r = 0
    while r < len(res):
        j = r + 1
        while j < len(res):
            if len(set(res[r]).intersection(res[j])) / (len(res[j])) >= 0.5:
                res[r] = list(set(res[r]).union(res[j]))
                res.pop(j)
            else:
                j = j + 1
        r = r + 1

    r = 0
    while r < len(res):
        j = r + 1
        while j < len(res):
            if len(set(res[r]).intersection(res[j])) / (len(res[j])) >= 0.33:
                res[r] = list(set(res[r]).union(res[j]))
                res.pop(j)
            else:
                j = j + 1
        r = r + 1

    r = 0
    while r < len(res):
        j = r + 1
        while j < len(res):
            if len(set(res[r]).intersection(res[j])) / (len(res[j])) >= 0.2:
                res[r] = list(set(res[r]).union(res[j]))
                res.pop(j)
            else:
                j = j + 1
        r = r + 1

    return res



import random
import math

"""
Xu, X., Yuruk, N., Feng, Z., & Schweiger, T. A. (2007, August). 
Scan: a structural clustering algorithm for networks. 
In Proceedings of the 13th ACM SIGKDD international conference on Knowledge discovery and data mining (pp. 824-833). .
"""


def cal_similarity(G, node_i, node_j):
    s1 = set(G.neighbors(node_i))
    s1.add(node_i)
    s2 = set(G.neighbors(node_j))
    s2.add(node_j)
    return len(s1 & s2) / math.sqrt(len(s1) * len(s2))


class SCAN_nx:
    def __init__(self, g, epsilon=0.5, mu=3):
        self.g = g
        self.epsilon = epsilon
        self.mu = mu

    def get_epsilon_neighbor(self, node):
        return [
            neighbor
            for neighbor in self.g.neighbors(node)
            if cal_similarity(self.g, node, neighbor) >= self.epsilon
        ]

    def is_core(self, node):
        return len(self.get_epsilon_neighbor(node)) >= self.mu

    def get_hubs_outliers(self, communities):
        other_nodes = set(self.g.nodes.keys())
        node_community = {}
        for i, c in enumerate(communities):
            for node in c:
                other_nodes.discard(node)
                node_community[node] = i
        hubs = []
        outliers = []
        for node in other_nodes:
            neighbors = self.g.neighbors(node)
            neighbor_community = set()
            for neighbor in neighbors:
                if neighbor in node_community:
                    neighbor_community.add(node_community[neighbor])
            if len(neighbor_community) > 1:
                hubs.append(node)
            else:
                outliers.append(node)
        return hubs, outliers

    def execute(self):
        # random scan nodes
        visit_sequence = list(self.g.nodes.keys())
        random.shuffle(visit_sequence)
        communities = []
        for node_name in visit_sequence:
            node = self.g.nodes[node_name]
            if node.get("classified"):
                continue
            if self.is_core(node_name):  # a new algorithms
                community = [node_name]
                communities.append(community)
                node["type"] = "core"
                node["classified"] = True
                queue = self.get_epsilon_neighbor(node_name)

                while len(queue) != 0:
                    temp = queue.pop(0)
                    if not self.g.nodes[temp].get("classified"):
                        self.g.nodes[temp]["classified"] = True
                        community.append(temp)
                    if not self.is_core(temp):
                        continue
                    R = self.get_epsilon_neighbor(temp)
                    for r in R:
                        node_r = self.g.nodes[r]
                        is_classified = node_r.get("classified")
                        if is_classified:
                            continue
                        node_r["classified"] = True
                        community.append(r)
                        if node_r.get("type") != "non-member":
                            queue.append(r)
                        else:
                            node["type"] = "non-member"
        return list(communities)
